/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2009 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: quad.c,v 1.170 2009/09/16 06:21:53 zlb Exp $ */

/* Numerical Quadrature */

#include "phg.h"
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifdef Length
# undef Length
#endif
#define Length(wts)	(sizeof(wts) / (sizeof(wts[0])))

typedef QUAD *(*QGEN_FUNC) (int dim, int order);

void
phgQuadFree(QUAD **quad)
{
    if ((*quad) == NULL)
	return;
    phgFree((*quad)->name);
    phgFree((*quad)->points);
    phgFree((*quad)->weights);
    phgFree((*quad));

    *quad = NULL;
}

#define nquad0	(sizeof(table0) / sizeof(*table0))
#define GetQuadrature(qgen_func)					\
    static int nquad;							\
    static QUAD **table = NULL;						\
    static BOOLEAN warned = FALSE;					\
									\
    if (order < 0) {							\
	/* free dynamically generated quadrature rules */		\
	if (table != NULL) {						\
	    int i, j;							\
	    for (i = nquad0; i < nquad; i++) {				\
		for (j = i + 1; j < nquad && table[j] == table[i]; j++)	\
		    table[j] = NULL;					\
		phgQuadFree(&table[i]);					\
	    }								\
	    phgFree(table);						\
	    table = NULL;						\
	}								\
	return NULL;							\
    }									\
    if (table == NULL) {						\
	/* initialization */						\
	nquad = nquad0;							\
	table = phgAlloc(sizeof(table0));				\
	memcpy(table, table0, sizeof(table0));				\
    }									\
    if (order >= nquad) {						\
	if (qgen_func == NULL) {					\
	    if (!warned && phgRank == 0) {				\
		phgWarning("%dD quad rule of order %d unavailable.\n",	\
				table[0]->dim, order);			\
		warned = TRUE;						\
	    }								\
	    order = nquad - 1;						\
	}								\
	else {								\
	    if (qgen_func == phgQuadTensorProductRule)			\
		order = (order / 2) * 2	+ 1;				\
	    table = phgRealloc_(table, (order + 1) * sizeof(*table),	\
					nquad * sizeof(*table));	\
	    memset(table + nquad, 0, (order - nquad + 1) * sizeof(*table)); \
	    nquad = order + 1;						\
	}								\
    }									\
    if (table[order] == NULL) {						\
	if (FALSE && table[0]->dim > 1 && phgRank == 0) {		\
	    phgWarning("using %dD-%d tensor product quadrature rule.\n",\
			table[0]->dim, order);				\
	}								\
	table[order] = (qgen_func)(table[0]->dim, order);		\
	if (order > nquad0 && qgen_func == phgQuadTensorProductRule)	\
	    table[order - 1] = table[order];				\
    }									\
    return table[order];

QUAD *
phgQuadGetQuad1D(int order)
{
    /* Table of static 1D quadrature rules */
    static QUAD *table0[] = { QUAD_1D_P1,
	QUAD_1D_P1,	QUAD_1D_P2,	QUAD_1D_P3,
	QUAD_1D_P4,	QUAD_1D_P5,	QUAD_1D_P6,
	QUAD_1D_P7,	QUAD_1D_P8,	QUAD_1D_P9,
	QUAD_1D_P10,	QUAD_1D_P11,	QUAD_1D_P12,
	QUAD_1D_P13,	QUAD_1D_P14,	QUAD_1D_P15,
	QUAD_1D_P16,	QUAD_1D_P17,	QUAD_1D_P18,
	QUAD_1D_P19,	QUAD_1D_P20,	QUAD_1D_P21
    };
    GetQuadrature(phgQuadTensorProductRule)
}

QUAD *
phgQuadGetQuad2D(int order)
{
    /* Table of static 2D quadrature rules */
    static QUAD *table0[] = { QUAD_2D_P1,
	QUAD_2D_P1,	QUAD_2D_P2,	QUAD_2D_P3,	QUAD_2D_P4,
	QUAD_2D_P5,	QUAD_2D_P6,	QUAD_2D_P7,	QUAD_2D_P8,
	QUAD_2D_P9,	QUAD_2D_P10,	QUAD_2D_P11,	QUAD_2D_P12,
	QUAD_2D_P13,	QUAD_2D_P14,	QUAD_2D_P15,	QUAD_2D_P16,
	QUAD_2D_P17,	QUAD_2D_P18,	QUAD_2D_P19,	QUAD_2D_P20,
	QUAD_2D_P21
    };
    GetQuadrature(phgQuadTensorProductRule)
}

QUAD *
phgQuadGetQuad3D(int order)
{
    /* Table of static 3D quadrature (cubature) rules */
    static QUAD *table0[] = { QUAD_3D_P1,
	QUAD_3D_P1,	QUAD_3D_P2,	QUAD_3D_P3,	QUAD_3D_P4,
	QUAD_3D_P5,	QUAD_3D_P6,	QUAD_3D_P7,	QUAD_3D_P8,
	QUAD_3D_P9,	QUAD_3D_P10,	QUAD_3D_P11,	QUAD_3D_P12,
	QUAD_3D_P13,	QUAD_3D_P14
    };
    GetQuadrature(phgQuadTensorProductRule)
}

/*--------------------------- 1D quadrature rules ------------------------*/

static FLOAT QUAD_1D_P1_wts[] = {
    Dup2(1.L)
};
static FLOAT QUAD_1D_P1_pts[Length(QUAD_1D_P1_wts) * 2] = {
    Perm2(.5L)
};
QUAD QUAD_1D_P1_ = {
    "1D P1",			/* name */
    1,				/* dim */
    1,				/* order */
    Length(QUAD_1D_P1_wts),	/* npoints */
    QUAD_1D_P1_pts,		/* points */
    QUAD_1D_P1_wts,		/* weights */
    -1				/* id */
};

static FLOAT QUAD_1D_P3_wts[] = {
    Dup11(.5L)
};
static FLOAT QUAD_1D_P3_pts[Length(QUAD_1D_P3_wts) * 2] = {
    /* (3 - sqrt(3)) / 6, (3 + sqrt(3)) / 6 */
    Perm11(.2113248654051871177454256097490212L)
};
QUAD QUAD_1D_P3_ = {
    "1D P3",			/* name */
    1,				/* dim */
    3,				/* order */
    Length(QUAD_1D_P3_wts),	/* npoints */
    QUAD_1D_P3_pts,		/* points */
    QUAD_1D_P3_wts,		/* weights */
    -1				/* id */
};

static FLOAT QUAD_1D_P5_wts[] = {
    Dup11(5.L / 18.L),
    Dup2(4.L / 9.L)
};
static FLOAT QUAD_1D_P5_pts[Length(QUAD_1D_P5_wts) * 2] = {
    /* (5 - sqrt(15)) / 10, 1 / 2, (5 + sqrt(15)) / 10 */
    Perm11(.1127016653792583114820734600217600L),
    Perm2(.5L)
};
QUAD QUAD_1D_P5_ = {
    "1D P5",			/* name */
    1,				/* dim */
    5,				/* order */
    Length(QUAD_1D_P5_wts),	/* npoints */
    QUAD_1D_P5_pts,		/* points */
    QUAD_1D_P5_wts,		/* weights */
    -1				/* id */
};

static FLOAT QUAD_1D_P7_wts[] = {
    /* (18 - sqrt(30)) / 72, (18 + sqrt(30)) / 72 */
    Dup11(.1739274225687269286865319746109997L),
    Dup11(.3260725774312730713134680253890002L)
};
static FLOAT QUAD_1D_P7_pts[Length(QUAD_1D_P7_wts) * 2] = {
    /* (35 \pm sqrt(525 \pm 70 * sqrt(30))) / 70 */
    Perm11(.0694318442029737123880267555535953L),
    Perm11(.3300094782075718675986671204483776L)
};
QUAD QUAD_1D_P7_ = {
    "1D P7",			/* name */
    1,				/* dim */
    7,				/* order */
    Length(QUAD_1D_P7_wts),	/* npoints */
    QUAD_1D_P7_pts,		/* points */
    QUAD_1D_P7_wts,		/* weights */
    -1				/* id */
};

static FLOAT QUAD_1D_P9_wts[] = {
    /* (322 \pm 13 * sqrt(70)) / 1800 */
    Dup2(128.L / 450.L),
    Dup11(.1184634425280945437571320203599586L),
    Dup11(.2393143352496832340206457574178190L)
};
static FLOAT QUAD_1D_P9_pts[Length(QUAD_1D_P9_wts) * 2] = {
    /* (21 \pm sqrt(245 \pm 14 * sqrt(70))) / 42 */
    Perm2(0.5L),
    Perm11(.0469100770306680036011865608503036L),
    Perm11(.2307653449471584544818427896498955L)
};
QUAD QUAD_1D_P9_ = {
    "1D P9",			/* name */
    1,				/* dim */
    9,				/* order */
    Length(QUAD_1D_P9_wts),	/* npoints */
    QUAD_1D_P9_pts,		/* points */
    QUAD_1D_P9_wts,		/* weights */
    -1				/* id */
};

static FLOAT QUAD_1D_P11_wts[] = {
    Dup11(.0856622461895851725201480710863664L),
    Dup11(.1803807865240693037849167569188581L),
    Dup11(.2339569672863455236949351719947755L)
};
static FLOAT QUAD_1D_P11_pts[Length(QUAD_1D_P11_wts) * 2] = {
    Perm11(.9662347571015760139061507772469973L),
    Perm11(.8306046932331322568306997975099527L),
    Perm11(.6193095930415984543152508608403560L)
};
QUAD QUAD_1D_P11_ = {
    "1D P11",			/* name */
    1,				/* dim */
    11,				/* order */
    Length(QUAD_1D_P11_wts),	/* npoints */
    QUAD_1D_P11_pts,		/* points */
    QUAD_1D_P11_wts,		/* weights */
    -1				/* id */
};

static FLOAT QUAD_1D_P13_wts[] = {
    Dup11(.0647424830844348466353057163395410L),
    Dup11(.1398526957446383339507338857118898L),
    Dup11(.1909150252525594724751848877444876L),
    Dup2(.2089795918367346938775510204081633L)
};
static FLOAT QUAD_1D_P13_pts[Length(QUAD_1D_P13_wts) * 2] = {
    Perm11(.9745539561713792622630948420239256L),
    Perm11(.8707655927996972199319323866403942L),
    Perm11(.7029225756886985834533032060384807L),
    Perm2(.5000000000000000000000000000000000L)
};
QUAD QUAD_1D_P13_ = {
    "1D P13",			/* name */
    1,				/* dim */
    13,				/* order */
    Length(QUAD_1D_P13_wts),	/* npoints */
    QUAD_1D_P13_pts,		/* points */
    QUAD_1D_P13_wts,		/* weights */
    -1				/* id */
};

static FLOAT QUAD_1D_P15_wts[] = {
    Dup11(.0506142681451881295762656771549811L),
    Dup11(.1111905172266872352721779972131204L),
    Dup11(.1568533229389436436689811009933007L),
    Dup11(.1813418916891809914825752246385978L)
};
static FLOAT QUAD_1D_P15_pts[Length(QUAD_1D_P15_wts) * 2] = {
    Perm11(.9801449282487681158417804342847365L),
    Perm11(.8983332387068133697957769682379152L),
    Perm11(.7627662049581644929088695245946232L),
    Perm11(.5917173212478249024697380711800920L)
};
QUAD QUAD_1D_P15_ = {
    "1D P15",		/* name */
    1,			/* dim */
    15,			/* order */
    Length(QUAD_1D_P15_wts),	/* npoints */
    QUAD_1D_P15_pts,	/* points */
    QUAD_1D_P15_wts,	/* weights */
    -1			/* id */
};

static FLOAT QUAD_1D_P17_wts[] = {
    Dup11(.0406371941807872059859460790552618L),
    Dup11(.0903240803474287020292360156214564L),
    Dup11(.1303053482014677311593714347093164L),
    Dup11(.1561735385200014200343152032922218L),
    Dup2(.1651196775006298815822625346434870L)
};
static FLOAT QUAD_1D_P17_pts[Length(QUAD_1D_P17_wts) * 2] = {
    Perm11(.9840801197538130449177881014518364L),
    Perm11(.9180155536633178971497148940348674L),
    Perm11(.8066857163502951986543510196707371L),
    Perm11(.6621267117019044645192690073216683L),
    Perm2(.5000000000000000000000000000000000L)
};
QUAD QUAD_1D_P17_ = {
    "1D P17",		/* name */
    1,			/* dim */
    17,			/* order */
    Length(QUAD_1D_P17_wts),	/* npoints */
    QUAD_1D_P17_pts,	/* points */
    QUAD_1D_P17_wts,	/* weights */
    -1			/* id */
};

static FLOAT QUAD_1D_P19_wts[] = {
    Dup11(.0333356721543440687967844049466659L),
    Dup11(.0747256745752902965728881698288487L),
    Dup11(.1095431812579910219977674671140816L),
    Dup11(.1346333596549981775456134607847347L),
    Dup11(.1477621123573764350869464973256692L)
};
static FLOAT QUAD_1D_P19_pts[Length(QUAD_1D_P19_wts) * 2] = {
    Perm11(.9869532642585858600389820060422260L),
    Perm11(.9325316833444922553660483442117465L),
    Perm11(.8397047841495122031171636825574368L),
    Perm11(.7166976970646235953996329715828921L),
    Perm11(.5744371694908156054424130005648600L)
};
QUAD QUAD_1D_P19_ = {
    "1D P19",		/* name */
    1,			/* dim */
    19,			/* order */
    Length(QUAD_1D_P19_wts),	/* npoints */
    QUAD_1D_P19_pts,	/* points */
    QUAD_1D_P19_wts,	/* weights */
    -1			/* id */
};

static FLOAT QUAD_1D_P21_wts[] = {
    Dup11(.0278342835580868332413768602212743L),
    Dup11(.0627901847324523123173471496119701L),
    Dup11(.0931451054638671257130488207158279L),
    Dup11(.1165968822959952399592618524215876L),
    Dup11(.1314022722551233310903444349452546L),
    Dup2(.1364625433889503153572417641681711L)
};
static FLOAT QUAD_1D_P21_pts[Length(QUAD_1D_P21_wts) * 2] = {
    Perm11(.9891143290730284964019690005614287L),
    Perm11(.9435312998840476495375788846519636L),
    Perm11(.8650760027870246620467081260155767L),
    Perm11(.7595480646034059079628628347293048L),
    Perm11(.6347715779761724861657659927004308L),
    Perm2(.5000000000000000000000000000000000L)
};
QUAD QUAD_1D_P21_ = {
    "1D P21",		/* name */
    1,			/* dim */
    21,			/* order */
    Length(QUAD_1D_P21_wts),	/* npoints */
    QUAD_1D_P21_pts,	/* points */
    QUAD_1D_P21_wts,	/* weights */
    -1			/* id */
};

/*--------------------------- 2D cubature rules ------------------------*/

static FLOAT QUAD_2D_P1_wts[] = {
    Dup3(1.000000000000000000000000000000000L)
};
static FLOAT QUAD_2D_P1_pts[Length(QUAD_2D_P1_wts) * 3] = {
    /* 1/3 */
    Perm3(.3333333333333333333333333333333333L)
};
QUAD QUAD_2D_P1_ = {
    "2D P1",			/* name */
    2,				/* dim */
    1,				/* order */
    Length(QUAD_2D_P1_wts),	/* npoints */
    QUAD_2D_P1_pts,		/* points */
    QUAD_2D_P1_wts,		/* weights */
    -1				/* id */
};

static FLOAT QUAD_2D_P2_wts[] = {
    /* 1/3 */
    Dup21(.3333333333333333333333333333333333L)
};
static FLOAT QUAD_2D_P2_pts[Length(QUAD_2D_P2_wts) * 3] = {
    /* 1/6 */
    Perm21(.1666666666666666666666666666666667L)
};
QUAD QUAD_2D_P2_ = {
    "2D P2",			/* name */
    2,				/* dim */
    2,				/* order */
    Length(QUAD_2D_P2_wts),	/* npoints = 3 */
    QUAD_2D_P2_pts,		/* points */
    QUAD_2D_P2_wts,		/* weights */
    -1				/* id */
};

#if 0
/* Note: this rule has points on the edges */
static FLOAT QUAD_2D_P3_wts[] = {
    /* 1/30 */
    Dup21(.0333333333333333333333333333333333L),
    /* 3/10 */
    Dup21(.3000000000000000000000000000000000L)
};
static FLOAT QUAD_2D_P3_pts[Length(QUAD_2D_P3_wts) * 3] = {
    Perm21(.5000000000000000000000000000000000L),
    /* 1/6 */
    Perm21(.1666666666666666666666666666666667L)
};
#else
static FLOAT QUAD_2D_P3_wts[] = {
    Dup21(.2811498024409796482535143227020770L),
    Dup21(.0521835308923536850798190106312564L)
};
static FLOAT QUAD_2D_P3_pts[Length(QUAD_2D_P3_wts) * 3] = {
    Perm21(.1628828503958919109001618041849063L),
    Perm21(.4779198835675637000000000000000000L)
};
#endif
QUAD QUAD_2D_P3_ = {
    "2D P3",			/* name */
    2,				/* dim */
    3,				/* order */
    Length(QUAD_2D_P3_wts),	/* npoints = 6 */
    QUAD_2D_P3_pts,		/* points */
    QUAD_2D_P3_wts,		/* weights */
    -1				/* id */
};

static FLOAT QUAD_2D_P4_wts[] = {
    /* (620 + sqrt(213125 - 53320 * sqrt(10))) / 3720 */
    Dup21(.2233815896780114656950070084331228L),
    /* (620 - sqrt(213125 - 53320 * sqrt(10))) / 3720 */
    Dup21(.1099517436553218676383263249002105L)
};
static FLOAT QUAD_2D_P4_pts[Length(QUAD_2D_P4_wts) * 3] = {
    /* (8 - sqrt(10) + sqrt(38 - 44 * sqrt(2 / 5))) / 18 */
    Perm21(.4459484909159648863183292538830519L),
    /* (8 - sqrt(10) - sqrt(38 - 44 * sqrt(2 / 5))) / 18 */
    Perm21(.0915762135097707434595714634022015L)
};
QUAD QUAD_2D_P4_ = {
    "2D P4",			/* name */
    2,				/* dim */
    4,				/* order */
    Length(QUAD_2D_P4_wts),	/* npoints = 6 */
    QUAD_2D_P4_pts,		/* points */
    QUAD_2D_P4_wts,		/* weights */
    -1				/* id */
};

static FLOAT QUAD_2D_P5_wts[] = {
    /* (155 - sqrt(15)) / 1200 */
    Dup21(.1259391805448271525956839455001813L),
    /* (155 + sqrt(15)) / 1200 */
    Dup21(.1323941527885061807376493878331519L),
    /* 9 / 40 */
    Dup3(.2250000000000000000000000000000000L)
};
static FLOAT QUAD_2D_P5_pts[Length(QUAD_2D_P5_wts) * 3] = {
    /* (6 - sqrt(15)) / 21 */
    Perm21(.1012865073234563388009873619151238L),
    /* (6 + sqrt(15)) / 21 */
    Perm21(.4701420641051150897704412095134476L),
    /* 1 / 3 */
    Perm3(.3333333333333333333333333333333333L)
};
QUAD QUAD_2D_P5_ = {
    "2D P5",			/* name */
    2,				/* dim */
    5,				/* order */
    Length(QUAD_2D_P5_wts),	/* npoints = 7 */
    QUAD_2D_P5_pts,		/* points */
    QUAD_2D_P5_wts,		/* weights */
    -1				/* id */
};

/* Note: an 11-point rule has been found in:
 *	 Day, David M., Mark A. Taylor, "A new 11 point degree 6 cubature
 *	 	formula for the triangle [Proceedings of the ICIAM 2007],"
 *	 Journal Article, Proceedings of Applied Mathematics and Mechanics,
 *	 Accepted/Published January 2008. */
static FLOAT QUAD_2D_P6_wts[] = {
    Dup21(.0508449063702068169209368091068690L),
    Dup21(.1167862757263793660252896113855794L),
    Dup111(.0828510756183735751935534564204425L)
};
static FLOAT QUAD_2D_P6_pts[Length(QUAD_2D_P6_wts) * 3] = {
    Perm21(.0630890144915022283403316028708192L),
    Perm21(.2492867451709104212916385531070191L),
    Perm111(.0531450498448169473532496716313981L,
	    .3103524510337844054166077339565522L)
};
QUAD QUAD_2D_P6_ = {
    "2D P6",			/* name */
    2,				/* dim */
    6,				/* order */
    Length(QUAD_2D_P6_wts),	/* npoints = 12 */
    QUAD_2D_P6_pts,		/* points */
    QUAD_2D_P6_wts,		/* weights */
    -1				/* id */
};

#if 0
static FLOAT QUAD_2D_P7_wts[] = {
    Dup21(0.0102558174092),
    Dup111(0.1116047046647),
    Dup21(0.1679775595335),
    Dup21(0.2652238803946)
};
static FLOAT QUAD_2D_P7_pts[Length(QUAD_2D_P7_wts) * 3] = {
    Perm21(0.),
    Perm111(0.7839656651012, 0.1738960507345),
    Perm21(0.4743880861752),
    Perm21(0.2385615300181)
};
#else
static FLOAT QUAD_2D_P7_wts[] = {
    Dup21(.0135338625156655615668230924525939L),
    Dup21(.0789512544320109813765214502977033L),
    Dup21(.1286079278189060745566555330895234L),
    Dup111(.0561201442833753579166666287467563L)
};
static FLOAT QUAD_2D_P7_pts[Length(QUAD_2D_P7_wts) * 3] = {
    Perm21(.0282639241560763402235960069132400L),
    Perm21(.4743113232672225752752252279318165L),
    Perm21(.2411433258498488102541435126703621L),
    Perm111(.7612227480245238000000000000000000L,
	    .0462708777988089106409255939170205L)
};
#endif
QUAD QUAD_2D_P7_ = {
    "2D P7",			/* name */
    2,				/* dim */
    7,				/* order */
    Length(QUAD_2D_P7_wts),	/* npoints = 15 */
    QUAD_2D_P7_pts,		/* points */
    QUAD_2D_P7_wts,		/* weights */
    -1				/* id */
};

static FLOAT QUAD_2D_P8_wts[] = {
    Dup3(.1443156076777871682510911104890646L),
    Dup21(.1032173705347182502817915502921290L),
    Dup21(.0324584976231980803109259283417806L),
    Dup21(.0950916342672846247938961043885843L),
    Dup111(.0272303141744349942648446900739089L)
};
static FLOAT QUAD_2D_P8_pts[Length(QUAD_2D_P8_wts) * 3] = {
    Perm3(.3333333333333333333333333333333333L),
    Perm21(.1705693077517602066222935014914645L),
    Perm21(.0505472283170309754584235505965989L),
    Perm21(.4592925882927231560288155144941693L),
    Perm111(.2631128296346381134217857862846436L,
	    .0083947774099576053372138345392944L)
};
QUAD QUAD_2D_P8_ = {
    "2D P8",			/* name */
    2,				/* dim */
    8,				/* order */
    Length(QUAD_2D_P8_wts),	/* npoints = 16 */
    QUAD_2D_P8_pts,		/* points */
    QUAD_2D_P8_wts,		/* weights */
    -1				/* id */
};

#if 0
static FLOAT QUAD_2D_P9_wts[] = {
    Dup21(0.0519871420646),
    Dup111(0.0707034101784),
    Dup111(0.0909390760952),
    Dup21(0.1032344051380),
    Dup21(0.1881601469167)
};
static FLOAT QUAD_2D_P9_pts[Length(QUAD_2D_P9_wts) * 3] = {
    Perm21(0.0451890097844),
    Perm111(0.7475124727339, 0.2220631655373),
    Perm111(0.1369912012649, 0.2182900709714),
    Perm21(0.4815198347833),
    Perm21(0.4036039798179)
};
QUAD QUAD_2D_P9_ = {
    "2D P9",			/* name */
    2,				/* dim */
    9,				/* order */
    21,				/* npoints */
    QUAD_2D_P9_pts,		/* points */
    QUAD_2D_P9_wts,		/* weights */
    -1				/* id */
};
#else
static FLOAT QUAD_2D_P9_wts[] = {
    Dup3(.0971357962827988338192419825072886L),
    Dup21(.0313347002271390705368548312872093L),
    Dup21(.0255776756586980312616787985589998L),
    Dup21(.0778275410047742793167393562994040L),
    Dup21(.0796477389272102530328917742640453L),
    Dup111(.0432835393772893772893772893772894L)
};
static FLOAT QUAD_2D_P9_pts[Length(QUAD_2D_P9_wts) * 3] = {
    Perm3(.3333333333333333333333333333333333L),
    Perm21(.4896825191987376277837069248361928L),
    Perm21(.0447295133944527098651065899662764L),
    Perm21(.4370895914929366372699303644353550L),
    Perm21(.1882035356190327302409612804673356L),
    Perm111(.7411985987844980206900798735234238L,
	    .2219629891607656956751025276931911L)
};
QUAD QUAD_2D_P9_ = {
    "2D P9",			/* name */
    2,				/* dim */
    9,				/* order */
    Length(QUAD_2D_P9_wts),	/* npoints = 19 */
    QUAD_2D_P9_pts,		/* points */
    QUAD_2D_P9_wts,		/* weights */
    -1				/* id */
};
#endif

static FLOAT QUAD_2D_P10_wts[] = {
    Dup3(.0809374287976228802571131238165019L),
    Dup21(.0772985880029631216825069823803434L),
    Dup21(.0784576386123717313680939208343967L),
    Dup21(.0174691679959294869176071632906781L),
    Dup21(.0042923741848328280304804020901319L),
    Dup111(.0374688582104676429790207654850445L),
    Dup111(.0269493525918799596454494795810967L)
};
static FLOAT QUAD_2D_P10_pts[Length(QUAD_2D_P10_wts) * 3] = {
    Perm3(.3333333333333333333333333333333333L),
    Perm21(.4272731788467755380904427175154472L),
    Perm21(.1830992224486750205215743848502200L),
    Perm21(.4904340197011305874539712223768484L),
    Perm21(.0125724455515805327313290850210413L),
    Perm111(.6542686679200661406665700955876279L,
	    .3080460016852477000000000000000000L),
    Perm111(.1228045770685592734301298174812812L,
	    .0333718337393047862408164417747804L)
};
QUAD QUAD_2D_P10_ = {
    "2D P10",			/* name */
    2,				/* dim */
    10,				/* order */
    Length(QUAD_2D_P10_wts),	/* npoints = 25 */
    QUAD_2D_P10_pts,		/* points */
    QUAD_2D_P10_wts,		/* weights */
    -1				/* id */
};

static FLOAT QUAD_2D_P11_wts[] = {
    Dup3(.0811779602968671595154759687498236L),
    Dup21(.0123240435069094941184739010162328L),
    Dup21(.0628280097444101072833394281602940L),
    Dup21(.0122203790493645297552122150039379L),
    Dup21(.0677013489528115099209888618232256L),
    Dup21(.0402196936288516904235668896075687L),
    Dup111(.0147622727177161013362930655877821L),
    Dup111(.0407279964582990396603369584816179L)
};
static FLOAT QUAD_2D_P11_pts[Length(QUAD_2D_P11_wts) * 3] = {
    Perm3(.3333333333333333333333333333333333L),
    Perm21(.0309383552454307848951950149913047L),
    Perm21(.4364981811341288419176152765599732L),
    Perm21(.4989847637025932662879869838313909L),
    Perm21(.2146881979585943366068758138782509L),
    Perm21(.1136831040421133902052931562283618L),
    Perm111(.8256187661648629043588062003083580L,
	    .1597423045918501898008607882250075L),
    Perm111(.6404723101348652676770365908189668L,
	    .3117837157095990000000000000000000L)
};
QUAD QUAD_2D_P11_ = {
    "2D P11",			/* name */
    2,				/* dim */
    11,				/* order */
    Length(QUAD_2D_P11_wts),	/* npoints = 28 */
    QUAD_2D_P11_pts,		/* points */
    QUAD_2D_P11_wts,		/* weights */
    -1				/* id */
};

static FLOAT QUAD_2D_P12_wts[] = {
    Dup21(.0061662610515590172338664837852304L),
    Dup21(.0628582242178851003542705130928825L),
    Dup21(.0347961129307089429893283972949994L),
    Dup21(.0436925445380384021354572625574750L),
    Dup21(.0257310664404553354177909230715644L),
    Dup111(.0223567732023034457118390767023200L),
    Dup111(.0173162311086588923716421008110341L),
    Dup111(.0403715577663809295178286992522368L)
};
static FLOAT QUAD_2D_P12_pts[Length(QUAD_2D_P12_wts) * 3] = {
    Perm21(.0213173504532103702468569755157282L),
    Perm21(.2712103850121159223459513403968947L),
    Perm21(.1275761455415859246738963251542836L),
    Perm21(.4397243922944602729797366234843611L),
    Perm21(.4882173897738048825646620652588110L),
    Perm111(.6958360867878034221416355232360725L,
	    .2813255809899395482481306929745527L),
    Perm111(.8580140335440726305905366166261782L,
	    .1162519159075971412413541478426018L),
    Perm111(.6089432357797878068561924377637101L,
	    .2757132696855141939747963460797640L)
};
QUAD QUAD_2D_P12_ = {
    "2D P12",			/* name */
    2,				/* dim */
    12,				/* order */
    Length(QUAD_2D_P12_wts),	/* npoints = 33 */
    QUAD_2D_P12_pts,		/* points */
    QUAD_2D_P12_wts,		/* weights */
    -1				/* id */
};

static FLOAT QUAD_2D_P13_wts[] = {
    Dup3(.0679600365868316442817744246808849L),
    Dup21(.0556019675304533287072574660104615L),
    Dup21(.0582784851191999814047670835133398L),
    Dup21(.0060523371035391718417928000322908L),
    Dup21(.0239944019288947307737107994509597L),
    Dup111(.0346412761408483704659868285109182L),
    Dup111(.0149654011051656672632458571329034L),
    Dup111(.0241790398115938191374457455730608L),
    Dup111(.0095906810035432627225950901661109L)
};
static FLOAT QUAD_2D_P13_pts[Length(QUAD_2D_P13_wts) * 3] = {
    Perm3(.3333333333333333333333333333333333L),
    Perm21(.4269414142598004060208125350313742L),
    Perm21(.2213722862918329006548125547050791L),
    Perm21(.0215096811088431838692913135340521L),
    Perm21(.4890769464525393499006897190902044L),
    Perm111(.6235459955536755708158543531862366L,
	    .3084417608921177746584718525412453L),
    Perm111(.8647077702954427753025459508956932L,
	    .1109220428034633954128695452216745L),
    Perm111(.7485071158999521951730185957887097L,
	    .1635974010678504802338879017109572L),
    Perm111(.7223577931241879652606201323047840L,
	    .2725158177734296661800504643540868L)
};
QUAD QUAD_2D_P13_ = {
    "2D P13",			/* name */
    2,				/* dim */
    13,				/* order */
    Length(QUAD_2D_P13_wts),	/* npoints = 37 */
    QUAD_2D_P13_pts,		/* points */
    QUAD_2D_P13_wts,		/* weights */
    -1				/* id */
};

static FLOAT QUAD_2D_P14_wts[] = {
    Dup3(.0585962852260285941278938063477560L),
    Dup21(.0017351512297252675680618638808094L),
    Dup21(.0261637825586145217778288591819783L),
    Dup21(.0039197292424018290965208275701454L),
    Dup21(.0122473597569408660972869899262505L),
    Dup21(.0281996285032579601073663071515657L),
    Dup21(.0508870871859594852960348275454540L),
    Dup21(.0504534399016035991910208971341189L),
    Dup111(.0170636442122334512900253993849472L),
    Dup111(.0096834664255066004075209630934194L),
    Dup111(.0363857559284850056220113277642717L),
    Dup111(.0069646633735184124253997225042413L)
};
static FLOAT QUAD_2D_P14_pts[Length(QUAD_2D_P14_wts) * 3] = {
    Perm3(.3333333333333333333333333333333333L),
    Perm21(.0099797608064584324152935295820524L),
    Perm21(.4799778935211883898105528650883899L),
    Perm21(.1538119591769669000000000000000000L),
    Perm21(.0740234771169878100000000000000000L),
    Perm21(.1303546825033300000000000000000000L),
    Perm21(.2306172260266531342996053700983831L),
    Perm21(.4223320834191478241144087137913939L),
    Perm111(.7862373859346610033296221140330900L,
	    .1906163600319009042461432828653034L),
    Perm111(.6305521436606074416224090755688129L,
	    .3623231377435471446183267343597729L),
    Perm111(.6265773298563063142335123137534265L,
	    .2907712058836674150248168174816732L),
    Perm111(.9142099849296254122399670993850469L,
	    .0711657108777507625475924502924336L)
};
QUAD QUAD_2D_P14_ = {
    "2D P14",			/* name */
    2,				/* dim */
    14,				/* order */
    Length(QUAD_2D_P14_wts),	/* npoints = 46 */
    QUAD_2D_P14_pts,		/* points */
    QUAD_2D_P14_wts,		/* weights */
    -1				/* id */
};

static FLOAT QUAD_2D_P15_wts[] = {
    Dup3(.0440387108784342798530173272149339L),
    Dup21(.0461847871820269799487156676019167L),
    Dup21(.0064989066173327165268828034928102L),
    Dup21(.0179936142526584032446699241671566L),
    Dup21(.0417731050391413541196860605641460L),
    Dup21(.0030595476091164665484301699283448L),
    Dup21(.0020124350525586473440903187565405L),
    Dup21(.0167756109305091223261114568879588L),
    Dup111(.0154607491897142748660880304092474L),
    Dup111(.0284998903395474233927395587533020L),
    Dup111(.0320943504834895956420992357370957L),
    Dup111(.0115085816368707112840232437732419L),
    Dup111(.0046143065289671031435871760918541L)
};
static FLOAT QUAD_2D_P15_pts[Length(QUAD_2D_P15_wts) * 3] = {
    Perm3(.3333333333333333333333333333333333L),
    Perm21(.2273322188191428742025043684922941L),
    Perm21(.4971625774318874298738098000160233L),
    Perm21(.4788497353489545833392292001438526L),
    Perm21(.4049860390982719916972446423426920L),
    Perm21(.0159312166717444321134277329412690L),
    Perm21(.1655832624260814000000000000000000L),
    Perm21(.0731336047192287277268738121073244L),
    Perm111(.6652607330722139390623644133856912L,
	    .3163528393449472300863381309502453L),
    Perm111(.7125219872425455330488490116233878L,
	    .0934607511499175300000000000000005L),
    Perm111(.5596483622353932184122484540192300L,
	    .3442290175821932000000000000000016L),
    Perm111(.8104765976190768630468327302905713L,
	    .1710472483142579515476503319255848L),
    Perm111(.9160756440317311885646088387783200L,
	    .0730559964791864896129490819274250L)
};
QUAD QUAD_2D_P15_ = {
    "2D P15",			/* name */
    2,				/* dim */
    15,				/* order */
    Length(QUAD_2D_P15_wts),	/* npoints = 52 */
    QUAD_2D_P15_pts,		/* points */
    QUAD_2D_P15_wts,		/* weights */
    -1				/* id */
};

static FLOAT QUAD_2D_P16_wts[] = {
    Dup3(.0480221886803770905518394045805199L),
    Dup21(.0147091003068019271034036428618692L),
    Dup21(.0295445865493192559953097267964641L),
    Dup21(.0261250173510883774985975654917156L),
    Dup21(.0027803873523900069750030161386621L),
    Dup21(.0318217730005366495034272900559496L),
    Dup21(.0086458343495096599011737341698489L),
    Dup111(.0143003329044953651466164253682521L),
    Dup111(.0278497772036008299522298734239535L),
    Dup111(.0070416734066360975623701880892807L),
    Dup111(.0178998382599337286017702090758108L),
    Dup111(.0274582003843497630724700381009172L),
    Dup111(.0072997969394317620841125440877777L)
};
static FLOAT QUAD_2D_P16_pts[Length(QUAD_2D_P16_wts) * 3] = {
    Perm3(.3333333333333333333333333333333333L),
    Perm21(.0817949831313738726414655931188610L),
    Perm21(.1653006019697796506267619329335566L),
    Perm21(.4685921053494613866946028972966056L),
    Perm21(.0144388134454166826141089566956602L),
    Perm21(.2417842853917833534068944592932077L),
    Perm21(.4953103429877699640654950868774055L),
    Perm111(.6505134026613522994311446848416867L,
	    .3313997445370895565813231681825939L),
    Perm111(.6040112814959970398494041030359670L,
	    .3032471627499421850415521780783469L),
    Perm111(.8021682575747416636168619478116671L,
	    .1880280595212371734441821142939888L),
    Perm111(.7565056064428283965511540757580608L,
	    .1835046685222968636823802774370004L),
    Perm111(.4659384387141181848838107335915464L,
	    .3596459487975046000000000000000100L),
    Perm111(.9063948439920415013624996618653400L,
	    .0771943712957554322825152250527139L)
};
QUAD QUAD_2D_P16_ = {
    "2D P16",			/* name */
    2,				/* dim */
    16,				/* order */
    Length(QUAD_2D_P16_wts),	/* npoints = 55 */
    QUAD_2D_P16_pts,		/* points */
    QUAD_2D_P16_wts,		/* weights */
    -1				/* id */
};

static FLOAT QUAD_2D_P17_wts[] = {
    Dup3(.0447568714443446293718364767042551L),
    Dup21(.0173668850267477964504911176477604L),
    Dup21(.0305993480761035327226656472689570L),
    Dup21(.0285877085785997802070400912176892L),
    Dup21(.0066474319297536932323184955454668L),
    Dup21(.0074761894020185118222455734708010L),
    Dup21(.0250499865038387453145589519055078L),
    Dup111(.0014798108921196449448095368275048L),
    Dup111(.0051211362467481060658943504057326L),
    Dup111(.0273173593695928059185316898423957L),
    Dup111(.0140057286759092815978663321140118L),
    Dup111(.0078092756974583600981098732328800L),
    Dup111(.0181657284597916721760720775172237L),
    Dup111(.0274443739924583277620834554147845L)
};
static FLOAT QUAD_2D_P17_pts[Length(QUAD_2D_P17_wts) * 3] = {
    Perm3(.3333333333333333333333333333333333L),
    Perm21(.0956985088627109399431625786023763L),
    Perm21(.1701386396787754467232472307956844L),
    Perm21(.4180206858679549762226346423934278L),
    Perm21(.4965814805066249549705403530622792L),
    Perm21(.0416621148288076427920788515983419L),
    Perm21(.4679329057294235782681900853617119L),
    Perm111(.9695311989037220561945405830595324L,
	    .0289250916202182460715280477140682L),
    Perm111(.7597243875386241295553271953226612L,
	    .2344417552635687745426605309129788L),
    Perm111(.2954993169683015000000000000000100L,
	    .4959112466607535754230345000437552L),
    Perm111(.6256063821576970270701920926669370L,
	    .3534176945414970676263249907709938L),
    Perm111(.8721744472331847929031830014156074L,
	    .1127286418142197686188888676807420L),
    Perm111(.7475123194400060400624067817608753L,
	    .1990702787978578813133914398155831L),
    Perm111(.5988687908832380598061676972635110L,
	    .3035851830713260765320205120458494L)
};
QUAD QUAD_2D_P17_ = {
    "2D P17",			/* name */
    2,				/* dim */
    17,				/* order */
    Length(QUAD_2D_P17_wts),	/* npoints = 61 */
    QUAD_2D_P17_pts,		/* points */
    QUAD_2D_P17_wts,		/* weights */
    -1				/* id */
};

static FLOAT QUAD_2D_P18_wts[] = {
    Dup21(.0139778616452860209795840079905549L),
    Dup21(.0005549069792132137850684555152509L),
    Dup21(.0210268138197046690284298685162450L),
    Dup21(.0340182121799276997472265274182211L),
    Dup111(.0279101658047749951418434740078169L),
    Dup111(.0182146861271508661267339566206858L),
    Dup111(.0142670236581097930775198241095567L),
    Dup111(.0142371230906750507043127637741560L),
    Dup111(.0192575838546747877991373836820213L),
    Dup111(.0097051322843806411487822763323902L),
    Dup111(.0076297881343321289957824556338534L),
    Dup111(.0106187391363503447944635436705283L),
    Dup111(.0057106698032758388134142143826895L),
    Dup111(.0043268574608764182945223447328327L)
};
static FLOAT QUAD_2D_P18_pts[Length(QUAD_2D_P18_wts) * 3] = {
    Perm21(.0732708864643828315786196714876895L),
    Perm21(.0039177489832282316427840744195806L),
    Perm21(.4675973189887110616515129966229624L),
    Perm21(.4179162109674113120121268105139935L),
    Perm111(.1653816933602894800544902692391766L,
	    .5636967056608707538051458939380737L),
    Perm111(.2875008944057839899961939131396606L,
	    .2860423261392047491209581074803029L),
    Perm111(.1258893143198247960170648399490380L,
	    .6960432186424611957925748602819539L),
    Perm111(.0632219159465026144935750801169980L,
	    .7605455518876824326145947637978687L),
    Perm111(.0789102274540205177520722103754889L,
	    .5920196312717585633226205754022254L),
    Perm111(.0380580535067857143261189915962621L,
	    .6836812596359998524801240874538131L),
    Perm111(.0142903521304540256499241103130749L,
	    .8517040371370558150285216534427664L),
    Perm111(.0129672723432531723123416343300903L,
	    .5747324928881490288994509386896897L),
    Perm111(.0076485948208408993307926288182273L,
	    .7355104408307292987031352244816406L),
    Perm111(.0127104605722554679311424918135822L,
	    .9393450876437317887074042026828225L)
};
QUAD QUAD_2D_P18_ = {
    "2D P18",			/* name */
    2,				/* dim */
    18,				/* order */
    Length(QUAD_2D_P18_wts),	/* npoints = 72 */
    QUAD_2D_P18_pts,		/* points */
    QUAD_2D_P18_wts,		/* weights */
    -1				/* id */
};

/* Note: the rule QUAD_2D_P19 was taken from the book by
 *	P. Solin, K. Segeth, and I. Dolezel,
 *	"Higer-order Finite Element Methods",
 *	Chapman and Hall/CRC Press, 2003. */
static FLOAT QUAD_2D_P19_wts[] = {
    Dup3(.0329063313889186520836143448464750L),
    Dup21(.0103307318912720533670399635717483L),
    Dup21(.0223872472630163925291845560351627L),
    Dup21(.0302661258694680708652801909825912L),
    Dup21(.0304909678021977810000315865785204L),
    Dup21(.0241592127416409049118480309866400L),
    Dup21(.0160508035868008752916227702764295L),
    Dup21(.0080845802617840604818056732421944L),
    Dup21(.0020793620274847807513475016743984L),
    Dup111(.0038848769049813897567049919927727L),
    Dup111(.0255741606120219038929297019526003L),
    Dup111(.0088809035733380577455259247035175L),
    Dup111(.0161245467617313912197852693278377L),
    Dup111(.0024919418174906754405846475759496L),
    Dup111(.0182428401189505783776657132097361L),
    Dup111(.0102585637361985213080480700423581L),
    Dup111(.0037999288553019139790731537136397L)
};
static FLOAT QUAD_2D_P19_pts[Length(QUAD_2D_P19_wts) * 3] = {
    Perm3(.3333333333333333333333333333333333L),
    Perm21(.4896099870730063319661310657482982L),
    Perm21(.4545368926978926620467593905357283L),
    Perm21(.4014166806494311873939956238106886L),
    Perm21(.2555516544030976113221817681092679L),
    Perm21(.1770779421521295516426752065159011L),
    Perm21(.1100610532279518613000849516773740L),
    Perm21(.0555286242518396712486784124713557L),
    Perm21(.0126218637772286684902347667787060L),
    Perm111(.6006337947946450000000000000000000L,
	    .3957547873569428623047946940658279L),
    Perm111(.1344667545307797856120431989326469L,
	    .5576032615887839683639532425011810L),
    Perm111(.7209870258173650552166529023382789L,
	    .2645669484065202080403017349012149L),
    Perm111(.5945270689558709246138892880265067L,
	    .3585393522059505884249269906459009L),
    Perm111(.8393314736808385786174900771484052L,
	    .1578074059685947447376736033595065L),
    Perm111(.2238614240979156913033693895065364L,
	    .7010879789261733673232883365595116L),
    Perm111(.8229313240698566316274715591605332L,
	    .1424216011133834373155747568772374L),
    Perm111(.9243442526207840294558591379015631L,
	    .0654946280829377033923265249859256L)
};
QUAD QUAD_2D_P19_ = {
    "2D P19",			/* name */
    2,				/* dim */
    19,				/* order */
    Length(QUAD_2D_P19_wts),	/* npoints = 73 */
    QUAD_2D_P19_pts,		/* points */
    QUAD_2D_P19_wts,		/* weights */
    -1				/* id */
};

static FLOAT QUAD_2D_P20_wts[] = {
    Dup3(.0125376079944966565735856367723948L),
    Dup21(.0274718698764242137484535496073598L),
    Dup21(.0097652722770514230413646914294237L),
    Dup21(.0013984195353918235239233631597867L),
    Dup21(.0092921026251851826304282034030330L),
    Dup21(.0165778760323669253260236250351840L),
    Dup111(.0206677623486650769614219700129729L),
    Dup111(.0208222355211545073068785561993297L),
    Dup111(.0095686384198490606888758450458320L),
    Dup111(.0244527709689724638856439207024089L),
    Dup111(.0031557306306305340038264003207296L),
    Dup111(.0121367963653212969370133090807574L),
    Dup111(.0149664801438864490365249118515707L),
    Dup111(.0063275933217777395693240327504398L),
    Dup111(.0013425603120636958849798512981433L),
    Dup111(.0027760769163475540677293561558015L),
    Dup111(.0107398444741849415551734474479517L),
    Dup111(.0053678057381874532052474100212697L)
};
static FLOAT QUAD_2D_P20_pts[Length(QUAD_2D_P20_wts) * 3] = {
    Perm3(.3333333333333333333333333333333333L),
    Perm21(.2158743059329919731902545438401828L),
    Perm21(.0753767665297472780972854309459163L),
    Perm21(.0103008281372217921136862160096969L),
    Perm21(.4936022112987001655119208321450536L),
    Perm21(.4615509381069252967410487102915180L),
    Perm111(.3286214064242369933034974609509133L,
	    .4293405702582103752139588004663984L),
    Perm111(.2604803617865687564195930170811535L,
	    .1015775342809694461687550061961797L),
    Perm111(.1370742358464553000000000000000000L,
	    .7100659730011301599879040745464079L),
    Perm111(.1467269458722997843041609884874530L,
	    .4985454776784148493896226967076119L),
    Perm111(.0269989777425532900000000000000000L,
	    .0491867226725820016197037125775872L),
    Perm111(.0618717859336170268417124700122339L,
	    .7796601465405693953603506190768108L),
    Perm111(.0477243674276219962083526801042934L,
	    .3704915391495476369201496202567388L),
    Perm111(.1206005151863643799672337870400794L,
	    .8633469487547526484979879960925217L),
    Perm111(.0026971477967097876716489145012827L,
	    .0561949381877455029878923019865887L),
    Perm111(.0030156332779423626572762598234710L,
	    .2086750067484213509575944630613577L),
    Perm111(.0299053757884570188069287738643386L,
	    .7211512409120340910281041502050941L),
    Perm111(.0067566542224609885399458175192278L,
	    .6400554419405418899040536682721647L)
};
QUAD QUAD_2D_P20_ = {
    "2D P20",                   /* name */
    2,                          /* dim */
    20,                         /* order */
    Length(QUAD_2D_P20_wts),    /* npoints = 88 */
    QUAD_2D_P20_pts,            /* points */
    QUAD_2D_P20_wts,            /* weights */
    -1                          /* id */
};

static FLOAT QUAD_2D_P21_wts[] = {
    Dup3(.0275622569528764809669070448245143L),
    Dup21(.0220602154134885011913507340331164L),
    Dup21(.0234600159386714884930134449523000L),
    Dup21(.0003268895950471905462145575015465L),
    Dup21(.0032653194629399682343353040958667L),
    Dup21(.0117564629154127977043079692133821L),
    Dup21(.0117807684199115168455575790986761L),
    Dup111(.0022688108188011408053357043343043L),
    Dup111(.0025960109644363200606737836654882L),
    Dup111(.0046345297858718602123478905615969L),
    Dup111(.0047943360545488579348574487199119L),
    Dup111(.0057124788367236115672506383429634L),
    Dup111(.0058658276043221216369557987000023L),
    Dup111(.0094137630590915875898182685203471L),
    Dup111(.0134149437966564249100220266108931L),
    Dup111(.0157169180920832459435000011378462L),
    Dup111(.0168636830144369045916509638861999L),
    Dup111(.0213900270853200983778322980803590L),
    Dup111(.0230767921894926813678808755218915L)
};
static FLOAT QUAD_2D_P21_pts[Length(QUAD_2D_P21_wts) * 3] = {
    Perm3(.3333333333333333333333333333333333L),
    Perm21(.2009352770650852798729618515641637L),
    Perm21(.4376591659619271797318338441880541L),
    Perm21(.0034339564905961768509599122096049L),
    Perm21(.0466434847753067534951762404321419L),
    Perm21(.3864222517630714909403520241677264L),
    Perm21(.0954354711085309101085716810414760L),
    Perm111(.9555138033504563605013147251467712L,
	    .0357186278731633582380416089754387L),
    Perm111(.8866388134288682261249005746914376L,
	    .1081432249156462115273886110463127L),
    Perm111(.7842628458804341542966439903981954L,
	    .2074644495998764568243804295157274L),
    Perm111(.8829239550502000327113489873168897L,
	    .0856847087203169400000000000000100L),
    Perm111(.6689919644410772404913224832098946L,
	    .3214940030142888168816832126834860L),
    Perm111(.5520721210355609641571609652527788L,
	    .4379422187933413835523680769629170L),
    Perm111(.7975929655965685676293142232957258L,
	    .1619164530635778567510067702038591L),
    Perm111(.6775147151197714846349911663441326L,
	    .2745047674019949038590029729073332L),
    Perm111(.5429974155890916053311361168391934L,
	    .4053359980750069279498908953763256L),
    Perm111(.7054599055699685616588563415406017L,
	    .1877376806564353427728167439451200L),
    Perm111(.5748005730665084622159824505498500L,
	    .3056968347660551665127925566498432L),
    Perm111(.4717788085046148166039770401349242L,
	    .3121444668708908816708046058155764L)
};
QUAD QUAD_2D_P21_ = {
    "2D P21",			/* name */
    2,				/* dim */
    21,				/* order */
    Length(QUAD_2D_P21_wts),	/* npoints = 91 */
    QUAD_2D_P21_pts,		/* points */
    QUAD_2D_P21_wts,		/* weights */
    -1				/* id */
};

/*---------------------------- 3D cubature rules ---------------------------*/

static FLOAT QUAD_3D_P1_wts[] = {
    Dup4(1.000000000000000000000000000000000L)
};
static FLOAT QUAD_3D_P1_pts[Length(QUAD_3D_P1_wts) * 4] = {
    Perm4(.2500000000000000000000000000000000L)
};
QUAD QUAD_3D_P1_ = {
    "3D P1",			/* name */
    3,				/* dim */
    1,				/* order */
    Length(QUAD_3D_P1_wts),	/* npoints = 1 */
    QUAD_3D_P1_pts,		/* points */
    QUAD_3D_P1_wts,		/* weights */
    -1				/* id */
};

static FLOAT QUAD_3D_P2_wts[] = {
    Dup31(.2500000000000000000000000000000000L)
};
static FLOAT QUAD_3D_P2_pts[Length(QUAD_3D_P2_wts) * 4] = {
    /* (5. - sqrt(5.)) / 20, (5. + 3 * sqrt(5.)) / 20 */
    Perm31(.1381966011250105151795413165634361L)
};
QUAD QUAD_3D_P2_ = {
    "3D P2",			/* name */
    3,				/* dim */
    2,				/* order */
    Length(QUAD_3D_P2_wts),	/* npoints = 4 */
    QUAD_3D_P2_pts,		/* points */
    QUAD_3D_P2_wts,		/* weights */
    -1				/* id */
};

#if 0
static FLOAT QUAD_3D_P3_wts[] = {
    Dup31(1.L / 40.L),
    Dup31(9.L / 40.L)
};
static FLOAT QUAD_3D_P3_pts[Length(QUAD_3D_P3_wts) * 4] = {
    Perm31(0.L, 1.L),
    Perm31(1.L / 3.L, 0.L)
};
#else
static FLOAT QUAD_3D_P3_wts[] = {
    /* 1 / 8 + sqrt((1715161837 - 406006699 * sqrt(17)) / 23101) / 3120 */
    Dup31(.1385279665118621423236176983756412L),
    /* 1 / 8 - sqrt((1715161837 - 406006699 * sqrt(17)) / 23101) / 3120 */
    Dup31(.1114720334881378576763823016243588L)
};
static FLOAT QUAD_3D_P3_pts[Length(QUAD_3D_P3_wts) * 4] = {
    /* (55 - 3 * sqrt(17) + sqrt(1022 - 134 * sqrt(17))) / 196 */
    Perm31(.3280546967114266473358058199811974L),
    /* (55 - 3 * sqrt(17) - sqrt(1022 - 134 * sqrt(17))) / 196 */
    Perm31(.1069522739329306827717020415706165L)
};
#endif
QUAD QUAD_3D_P3_ = {
    "3D P3",			/* name */
    3,				/* dim */
    3,				/* order */
    Length(QUAD_3D_P3_wts),	/* npoints = 8 */
    QUAD_3D_P3_pts,		/* points */
    QUAD_3D_P3_wts,		/* weights */
    -1				/* id */
};

#if 0
static FLOAT QUAD_3D_P4_wts[] = {
    Dup4(-148.L / 1875.L),	/* negative weight */
    Dup31(343.L / 7500.L),
    Dup22(56.L / 375.L)
};
static FLOAT QUAD_3D_P4_pts[Length(QUAD_3D_P4_wts) * 4] = {
    Perm4(0.25L),
    Perm31(1.L / 14.L),
    Perm22(0.1005964238332008)
};
QUAD QUAD_3D_P4_ = {
    "3D P4",			/* name */
    3,				/* dim */
    4,				/* order */
    11,				/* npoints */
    QUAD_3D_P4_pts,		/* points */
    QUAD_3D_P4_wts,		/* weights */
    -1				/* id */
};
#else
static FLOAT QUAD_3D_P4_wts[] = {
    Dup31(.0734930431163619493435869458636788L),
    Dup31(.1126879257180158503650149284763889L),
    Dup22(.0425460207770814668609320837732882L)
};
static FLOAT QUAD_3D_P4_pts[Length(QUAD_3D_P4_wts) * 4] = {
    Perm31(.0927352503108912262865589206603214L),
    Perm31(.3108859192633006097581474949404033L),
    Perm22(.0455037041256496500000000000000000L)
};
QUAD QUAD_3D_P4_ = {
    "3D P4",			/* name */
    3,				/* dim */
    4,				/* order */
    Length(QUAD_3D_P4_wts),	/* npoints = 14 */
    QUAD_3D_P4_pts,		/* points */
    QUAD_3D_P4_wts,		/* weights */
    -1				/* id */
};
#endif

#if 0
/* Stroud T3:5-1 p315 */
static FLOAT QUAD_3D_P5_wts[] = {
    Dup4(16.L / 135.L),
    /* (2665 + 14 * sqrt(15)) / 37800 */
    Dup31(.07193708377901862),
    /* (2665 - 14 * sqrt(15)) / 37800 */
    Dup31(.06906820722627239),
    Dup22(20.L / 378.L)
};
static FLOAT QUAD_3D_P5_pts[Length(QUAD_3D_P5_wts) * 4] = {
    Perm4(.25L),
    /* (7 - sqrt(15)) / 34, (13 + 3 * sqrt(15)) / 34 */
    Perm31(.09197107805272303),
    /* (7 + sqrt(15)) / 34, (13 - 3 * sqrt(15)) / 34 */
    Perm31(.31979362782962991),
    /* (10 - 2 * sqrt(15)) / 40, (10 + 2 * sqrt(15)) / 40 */
    Perm22(.05635083268962916)
};
QUAD QUAD_3D_P5_ = {
    "3D P5",			/* name */
    3,				/* dim */
    5,				/* order */
    15,				/* npoints */
    QUAD_3D_P5_pts,		/* points */
    QUAD_3D_P5_wts,		/* weights */
    -1				/* id */
};
#else
static FLOAT QUAD_3D_P5_wts[] = {
    Dup31(.1126879257180158507991856523332863L),
    Dup31(.0734930431163619495437102054863275L),
    Dup22(.0425460207770814664380694281202574L)

};
static FLOAT QUAD_3D_P5_pts[Length(QUAD_3D_P5_wts) * 4] = {
    Perm31(.3108859192633006097973457337634578L),
    Perm31(.0927352503108912264023239137370306L),
    Perm22(.0455037041256496494918805262793394L)
};
QUAD QUAD_3D_P5_ = {
    "3D P5",			/* name */
    3,				/* dim */
    5,				/* order */
    Length(QUAD_3D_P5_wts),	/* npoints = 14 */
    QUAD_3D_P5_pts,		/* points */
    QUAD_3D_P5_wts,		/* weights */
    -1				/* id */
};
#endif

static FLOAT QUAD_3D_P6_wts[] = {
    Dup31(.0399227502581674920996906275574800L),
    Dup31(.0100772110553206429480132374459369L),
    Dup31(.0553571815436547220951532778537260L),
    /* 27 / 560 */
    Dup211(.0482142857142857142857142857142857L)
};
static FLOAT QUAD_3D_P6_pts[Length(QUAD_3D_P6_wts) * 4] = {
    Perm31(.2146028712591520292888392193862850L),
    Perm31(.0406739585346113531155794489564101L),
    Perm31(.3223378901422755103439944707624921L),
    /* (3 - sqrt(5)) / 12, (5 + sqrt(5)) / 12, (1 + sqrt(5)) / 12 */
    Perm211(.0636610018750175252992355276057270L,
	    .6030056647916491413674311390609397L)
};
QUAD QUAD_3D_P6_ = {
    "3D P6",			/* name */
    3,				/* dim */
    6,				/* order */
    Length(QUAD_3D_P6_wts),	/* npoints = 24 */
    QUAD_3D_P6_pts,		/* points */
    QUAD_3D_P6_wts,		/* weights */
    -1				/* id */
};

static FLOAT QUAD_3D_P7_wts[] = {
    Dup4(.0954852894641308488605784361172264L),
    Dup31(.0423295812099670290762861707985467L),
    Dup22(.0318969278328575799342748240829425L),
    Dup211(.0372071307283346213696155611914811L),
    Dup211(.0081107708299033415661034334910965L)
};
static FLOAT QUAD_3D_P7_pts[Length(QUAD_3D_P7_wts) * 4] = {
    Perm4(.2500000000000000000000000000000000L),
    Perm31(.3157011497782027994234299995933115L),
    Perm22(.0504898225983963687630538229865625L),
    Perm211(.1888338310260010477364311038545858L,
	    .5751716375870000234832415770223075L),
    Perm211(.0212654725414832459888361014998199L,
	    .8108302410985485611181053798482324L)
};
QUAD QUAD_3D_P7_ = {
    "3D P7",                    /* name */
    3,                          /* dim */
    7,                          /* order */
    Length(QUAD_3D_P7_wts),     /* npoints = 35 */
    QUAD_3D_P7_pts,             /* points */
    QUAD_3D_P7_wts,             /* weights */
    -1                          /* id */
};

static FLOAT QUAD_3D_P8_wts[] = {
    Dup31(.0063971477799023213214514203351730L),
    Dup31(.0401904480209661724881611584798178L),
    Dup31(.0243079755047703211748691087719226L),
    Dup31(.0548588924136974404669241239903914L),
    Dup22(.0357196122340991824649509689966176L),
    Dup211(.0071831906978525394094511052198038L),
    Dup211(.0163721819453191175409381397561191L)
};
static FLOAT QUAD_3D_P8_pts[Length(QUAD_3D_P8_wts) * 4] = {
    Perm31(.0396754230703899012650713295393895L),
    Perm31(.3144878006980963137841605626971483L),
    Perm31(.1019866930627033000000000000000000L),
    Perm31(.1842036969491915122759464173489092L),
    Perm22(.0634362877545398924051412387018983L),
    Perm211(.0216901620677280048026624826249302L,
	    .7199319220394659358894349533527348L),
    Perm211(.2044800806367957142413355748727453L,
	    .5805771901288092241753981713906204L)
};
QUAD QUAD_3D_P8_ = {
    "3D P8",			/* name */
    3,				/* dim */
    8,				/* order */
    Length(QUAD_3D_P8_wts),	/* npoints = 46 */
    QUAD_3D_P8_pts,		/* points */
    QUAD_3D_P8_wts,		/* weights */
    -1				/* id */
};

static FLOAT QUAD_3D_P9_wts[] = {
    Dup4(.0564266931795062065887150432761254L),
    Dup31(.0033410950747134804029997443047177L),
    Dup31(.0301137547687737639073142384315749L),
    Dup31(.0064909609200615346357621168945686L),
    Dup211(.0098092858682545864319687425925550L),
    Dup211(.0281191538233654725516326174252926L),
    Dup211(.0078945869083315007683414920096088L),
    Dup211(.0194928120472399967169721944892460L)
};
static FLOAT QUAD_3D_P9_pts[Length(QUAD_3D_P9_wts) * 4] = {
    Perm4(.2500000000000000000000000000000000L),
    Perm31(.0340221770010448664654037088787676L),
    Perm31(.3227703335338005253913766832549640L),
    Perm31(.0604570774257749300000000000000000L),
    Perm211(.4553629909472082118003081504416430L,
	    .0056831773653301799061001601457447L),
    Perm211(.1195022553938258009779737046961144L,
	    .4631168324784899409762244936577296L),
    Perm211(.0280219557834011581550575066541237L,
	    .7252060768398674887385659542848099L),
    Perm211(.1748330320115746157853246459722452L,
	    .6166825717812564045706830909795407L)
};
QUAD QUAD_3D_P9_ = {
    "3D P9",			/* name */
    3,				/* dim */
    9,				/* order */
    Length(QUAD_3D_P9_wts),	/* npoints = 61 */
    QUAD_3D_P9_pts,		/* points */
    QUAD_3D_P9_wts,		/* weights */
    -1				/* id */
};

static FLOAT QUAD_3D_P10_wts[] = {
    Dup4(.0473997735560207383847388211780511L),
    Dup31(.0269370599922686998027641610048821L),
    Dup31(.0098691597167933832345577354301731L),
    Dup211(.0003619443443392536242398783848085L),
    Dup211(.0101358716797557927885164701150168L),
    Dup211(.0113938812201952316236209348807143L),
    Dup211(.0065761472770359041674557402004507L),
    Dup211(.0257397319804560712790360122596547L),
    Dup211(.0129070357988619906392954302494990L)
};
static FLOAT QUAD_3D_P10_pts[Length(QUAD_3D_P10_wts) * 4] = {
    Perm4(.2500000000000000000000000000000000L),
    Perm31(.3122500686951886477298083186868275L),
    Perm31(.1143096538573461505873711976536504L),
    Perm211(.0061380088247907478475937132484154L,
	    .9429887673452048661976305869182508L),
    Perm211(.0327794682164426707747210203323242L,
	    .3401847940871076327889879249496713L),
    Perm211(.4104307392189654942878978442515117L,
	    .1654860256196110516044901244445264L),
    Perm211(.0324852815648230478355149399784262L,
	    .1338521522120095130978284359645666L),
    Perm211(.1210501811455894259938950015950505L,
	    .4771903799042803505441064082969072L),
    Perm211(.1749793421839390242849492265283104L,
	    .6280718454753660106932760722179097L)
};
QUAD QUAD_3D_P10_ = {
    "3D P10",			/* name */
    3,				/* dim */
    10,				/* order */
    Length(QUAD_3D_P10_wts),	/* npoints = 81 */
    QUAD_3D_P10_pts,		/* points */
    QUAD_3D_P10_wts,		/* weights */
    -1				/* id */
};

static FLOAT QUAD_3D_P11_wts[] = {
    Dup4(.0394321080286588635073303344912044L),
    Dup31(.0156621262272791131500885627687651L),
    Dup31(.0033321723749014081444092361540149L),
    Dup31(.0140260774074897474374913609976924L),
    Dup211(.0010859075293324663068220983772355L),
    Dup211(.0202359604306631789111165731654084L),
    Dup211(.0117902148721258635368493804677018L),
    Dup211(.0076903149825212959011315780207389L),
    Dup211(.0044373057034592039047307260214396L),
    Dup211(.0114295484671840404107705525985940L),
    Dup1111(.0061856401712178114128192550838953L)
};
static FLOAT QUAD_3D_P11_pts[Length(QUAD_3D_P11_wts) * 4] = {
    Perm4(.2500000000000000000000000000000000L),
    Perm31(.1214913677765337944977023099080722L),
    Perm31(.0323162591510728963539544520895810L),
    Perm31(.3249261497886067978128419024144220L),
    Perm211(.0041483569716600120000000000000100L,
	    .5982659967901863502054538427761778L),
    Perm211(.2246246106763771414144751511649864L,
	    .4736622878323495714083696692020524L),
    Perm211(.0519050877725656967442272164426589L,
	    .5631447779082798987371019763030571L),
    Perm211(.1349301312162402042237591723429930L,
	    .7083588307858189538569950051271300L),
    Perm211(.0251911921082524729200511850653055L,
	    .7837195073400773754305740342999090L),
    Perm211(.3653187797817336139693319800988672L,
	    .1346039083168658000000000000000100L),
    Perm1111(.5229075395099384729652169275860292L,
	     .1407536305436959018425391394912785L,
	     .0097624381964526155082922803899778L)
};
QUAD QUAD_3D_P11_ = {
    "3D P11",			/* name */
    3,				/* dim */
    11,				/* order */
    Length(QUAD_3D_P11_wts),	/* npoints = 109 */
    QUAD_3D_P11_pts,		/* points */
    QUAD_3D_P11_wts,		/* weights */
    -1				/* id */
};

static FLOAT QUAD_3D_P12_wts[] = {
    Dup31(.0127676377009707415020377859651250L),
    Dup31(.0161211042379092682185815448957576L),
    Dup31(.0003716126985784422000425581898608L),
    Dup31(.0197174417866854576395533090381887L),
    Dup31(.0025713909308627183621823475944855L),
    Dup22(.0038172478705105759057531841278333L),
    Dup22(.0120872270776631131786031841931461L),
    Dup211(.0031058611584347334343168814992962L),
    Dup211(.0054595313364710306691274212676944L),
    Dup211(.0021428997484969975066685209365595L),
    Dup211(.0055246714672578296224493009816508L),
    Dup211(.0085369566944991804298517783667220L),
    Dup1111(.0115101778483233069733364412340329L),
    Dup1111(.0052038786528856136039679242125245L)
};
static FLOAT QUAD_3D_P12_pts[Length(QUAD_3D_P12_wts) * 4] = {
    Perm31(.1152997443514801453045572073891591L),
    Perm31(.2023362822405909000000000000000100L),
    Perm31(.0117175979576199515124790675483140L),
    Perm31(.3133064413678010672776027996445893L),
    Perm31(.2500057301155837000000000000000100L),
    Perm22(.0209954743507580066902018252705902L),
    Perm22(.1517740182474501000000000000000100L),
    Perm211(.0244197787434353647831400090476166L,
	    .8483292846978728506452088674348157L),
    Perm211(.2562070985320183089638201070856221L,
	    .4824873738738488478028928967297354L),
    Perm211(.0167903209796029906147179602885794L,
	    .6947719423657559269594985098841772L),
    Perm211(.1261608211398720423997070384689592L,
	    .7254104893029481189748595052126338L),
    Perm211(.4314351745263798472167069506637196L,
	    .1127219398928524152095997721100754L),
    Perm1111(.5016700624625056974751550716847613L,
	     .2724718028695223917835104675306045L,
	     .0720743288072989146501594845633582L),
    Perm1111(.2616448545378187456694550500639680L,
	     .0862922919470617319174235194435249L,
	     .0205654106558761383006248976271090L)
};
QUAD QUAD_3D_P12_ = {
    "3D P12",			/* name */
    3,				/* dim */
    12,				/* order */
    Length(QUAD_3D_P12_wts),	/* npoints = 140 */
    QUAD_3D_P12_pts,		/* points */
    QUAD_3D_P12_wts,		/* weights */
    -1				/* id */
};

static FLOAT QUAD_3D_P13_wts[] = {
    Dup4(.0150136877730831467506297063161598L),
    Dup31(.0182252092801734253237906894149010L),
    Dup31(.0070061092177414642403851869392631L),
    Dup22(.0164235497439495482954057310790553L),
    Dup22(.0051206100963605970726259694970217L),
    Dup22(.0111966986529049163438203208635196L),
    Dup211(.0156191497333799540095381130243197L),
    Dup211(.0024844230133164744190405677633847L),
    Dup211(.0016385985348182389384452530944075L),
    Dup211(.0059030304401249219717191465553587L),
    Dup211(.0110220824582180524044509798920153L),
    Dup211(.0004064518399641782258515551275585L),
    Dup1111(.0026879699729685420974578192665173L),
    Dup1111(.0019795048055267119053189467551074L),
    Dup1111(.0054463191814257912094318704010867L)
};
static FLOAT QUAD_3D_P13_pts[Length(QUAD_3D_P13_wts) * 4] = {
    Perm4(.2500000000000000000000000000000000L),
    Perm31(.1552160935190895031411578433570474L),
    Perm31(.3301226633396736002443319259519678L),
    Perm22(.1668064038938624992893778260114423L),
    Perm22(.0249237885477736177970140037486009L),
    Perm22(.0971976299157510014307224371624082L),
    Perm211(.2478592901573625669274691062082793L,
	    .4336532423568514471872606143476738L),
    Perm211(.0222315960818670029087952186089293L,
	    .8369003204037340051450948659569859L),
    Perm211(.1072786933130534104915045963958480L,
	    .7749803059750018075658787727417929L),
    Perm211(.1981768438839898114233184058214276L,
	    .5875693057822053025917201790359592L),
    Perm211(.0691792434773793164773253434746550L,
	    .6042000666600664470793526487111530L),
    Perm211(.0231147194719331600000000000000100L,
	    .9308757927924442486492022888288831L),
    Perm1111(.1178892875101960892229011747064425L,
	     .1165153642254072000000000000000100L,
	     .0420240011255154209567663430372000L),
    Perm1111(.6770327986022842635503222132674659L,
	     .0461653760246197108345804112217608L,
	     .0008443403189050397572989969213590L),
    Perm1111(.4848900886736331220108009415479083L,
	     .3588829429552020157242364690942109L,
	     .1381828349176287299695508090791236L)
};
QUAD QUAD_3D_P13_ = {
    "3D P13",			/* name */
    3,				/* dim */
    13,				/* order */
    Length(QUAD_3D_P13_wts),	/* npoints = 171 */
    QUAD_3D_P13_pts,		/* points */
    QUAD_3D_P13_wts,		/* weights */
    -1				/* id */
};

static FLOAT QUAD_3D_P14_wts[] = {
    Dup31(.0040651136652707670436208836835636L),
    Dup31(.0022145385334455781437599569500071L),
    Dup31(.0058134382678884505495373338821455L),
    Dup31(.0196255433858357215975623333961715L),
    Dup31(.0003875737905908214364538721248394L),
    Dup211(.0116429719721770369855213401005552L),
    Dup211(.0052890429882817131317736883052856L),
    Dup211(.0018310854163600559376697823488069L),
    Dup211(.0082496473772146452067449669173660L),
    Dup1111(.0030099245347082451376888748208987L),
    Dup1111(.0008047165617367534636261808760312L),
    Dup1111(.0029850412588493071187655692883922L),
    Dup1111(.0056896002418760766963361477811973L),
    Dup1111(.0041590865878545715670013980182613L),
    Dup1111(.0007282389204572724356136429745654L),
    Dup1111(.0054326500769958248216242340651926L)
};
static FLOAT QUAD_3D_P14_pts[Length(QUAD_3D_P14_wts) * 4] = {
    Perm31(.3272533625238485639093096692685289L),
    Perm31(.0447613044666850808837942096478842L),
    Perm31(.0861403311024363536537208740298857L),
    Perm31(.2087626425004322968265357083976176L),
    Perm31(.0141049738029209600635879152102928L),
    Perm211(.1021653241807768123476692526982584L,
	    .5739463675943338202814002893460107L),
    Perm211(.4075700516600107157213295651301783L,
	    .0922278701390201300000000000000000L),
    Perm211(.0156640007402803585557586709578084L,
	    .7012810959589440327139967673208426L),
    Perm211(.2254963562525029053780724154201103L,
	    .4769063974420887115860583354107011L),
    Perm1111(.3905984281281458000000000000000000L,
	     .2013590544123922168123077327235092L,
	     .0161122880710300298578026931548371L),
    Perm1111(.1061350679989021455556139029848079L,
	     .0327358186817269284944004077912660L,
	     .0035979076537271666907971523385925L),
    Perm1111(.5636383731697743896896816630648502L,
	     .2302920722300657454502526874135652L,
	     .1907199341743551862712487790637898L),
    Perm1111(.3676255095325860844092206775991167L,
	     .2078851380230044950717102125250735L,
	     .3312104885193449000000000000000000L),
    Perm1111(.7192323689817295295023401840796991L,
	     .1763279118019329762157993033636973L,
	     .0207602362571310090754973440611644L),
    Perm1111(.5278249952152987298409240075817276L,
	     .4372890892203418165526238760841918L,
	     .0092201651856641949463177554949220L),
    Perm1111(.5483674544948190728994910505607746L,
	     .3447815506171641228703671870920331L,
	     .0867217283322215394629438740085828L)
};
QUAD QUAD_3D_P14_ = {
    "3D P14",			/* name */
    3,				/* dim */
    14,				/* order */
    Length(QUAD_3D_P14_wts),	/* npoints = 236 */
    QUAD_3D_P14_pts,		/* points */
    QUAD_3D_P14_wts,		/* weights */
    -1				/* id */
};

/*------------------ functions computing order of a basis -------------------*/

#if 1
# define BasisOrder(u, e, i) (!DofIsHP(u) ? (u)->type->order : \
	(u)->hp->info->types[(u)->hp->elem_order[e->index]]->order)
#else	/* 0|1 */
# define BasisOrder(u, e, i) basis_order(u, e, i)

static inline int
basis_order(DOF *u, SIMPLEX *e, int ibas)
{
    DOF_TYPE *type;
    HP_TYPE *hp;
    BYTE *orders;
    GTYPE gtype;
    int n, m, gind, bind;

    if ((type = u->type) != NULL) {
	/* non h-p DOF */
	if (type->base_type != NULL)
	    type = type->base_type;
	if ((orders = type->orders) == NULL || ibas < 0 || ibas >= type->nbas)
	    return type->order;
	if (ibas < (n = NVert * type->np_vert))
	    return orders[ibas % type->np_vert];
	orders += type->np_vert;
	if ((ibas -= n) < (m = NEdge * type->np_edge))
	    return orders[ibas % type->np_edge];
	orders += type->np_edge;
	if ((ibas -= m) < (n = NFace * type->np_face))
	    return orders[ibas % type->np_face];
	return orders[type->np_face + ibas - n];
    }

    /* h-p DOF */
    hp = u->hp;
    phgDofGetElementBasisInfo(u, e, ibas, &gtype, &gind, &bind);
    switch (gtype) {
	case VERTEX:
	    break;
	case EDGE:
	    type = hp->info->types[hp->edge_order[e->edges[gind]]];
	    if (type->base_type != NULL)
		type = type->base_type;
	    if ((orders = type->orders) == NULL)
		return type->order;
	    return orders[type->np_vert + bind];
	case FACE:
	    type = hp->info->types[hp->face_order[e->faces[gind]]];
	    if (type->base_type != NULL)
		type = type->base_type;
	    if ((orders = type->orders) == NULL)
		return type->order;
	    return orders[type->np_vert + type->np_edge + bind];
	case ELEMENT:
	    type = hp->info->types[hp->elem_order[e->index]];
	    if (type->base_type != NULL)
		type = type->base_type;
	    if ((orders = type->orders) == NULL)
		return type->order;
	    return orders[type->np_vert + type->np_edge + type->np_face + bind];
	default:
	    phgError(1, "unexpected error.\n");
    }

    return 0;	/* make gcc happy */
}
#endif	/* 0|1 */

/*---------------- functions managing caches and quad_list ------------------*/

static QUAD **quad_list = NULL;
static size_t quad_list_count = 0, quad_list_allocated = 0;

#if 0
static inline QUAD_CACHE *
#else
static QUAD_CACHE *
#endif
get_cache(void **clist_ptr, QUAD *quad)
/* returns the cache in '**clist' matching 'quad' */
{
    QUAD_CACHE_LIST *clist;
    QUAD_CACHE *cache;

    if (quad->id < 0) {
	/* assign an id to quad */
	if (quad_list_count >= quad_list_allocated) {
	    quad_list = phgRealloc_(quad_list,
				(quad_list_allocated + 8) * sizeof(*quad_list),
				quad_list_allocated * sizeof(*quad_list));
	    quad_list_allocated += 8;
	}
	quad_list[quad_list_count] = quad;
	quad->id = quad_list_count++;
    }

    assert(quad->id < quad_list_count && quad_list[quad->id] == quad);

    if ((clist = *clist_ptr) == NULL) {
	clist = phgAlloc(sizeof(*clist));
	clist->n = 0;
	clist->caches = NULL;
	*clist_ptr = clist;
    }

    if (clist->n <= quad->id) {
	/* expand the list to quad->id + 1 entries */
	clist->caches = phgRealloc_(clist->caches,
				   (quad->id + 1) * sizeof(*cache),
				   clist->n * sizeof(*cache));
	while (clist->n <= quad->id)
	    clist->caches[clist->n++] = NULL;
    }

    if ((cache = clist->caches[quad->id]) == NULL) {
	cache = clist->caches[quad->id] = phgAlloc(sizeof(*cache));
	cache->e = NULL;
	cache->data = NULL;
    }

    return cache;
}

void
phgQuadReset(void)
{
    int i;

    /* Free quad_list entries */
    for (i = 0; i < quad_list_count; i++) {
	quad_list[i]->id = -1;
    }
    phgFree(quad_list);
    quad_list = NULL;
    quad_list_count = quad_list_allocated = 0;

    /* The calls below will free dynamically generated quadrature rules */
    phgQuadGetQuad1D(-1);
    phgQuadGetQuad2D(-1);
    phgQuadGetQuad3D(-1);
}

void
phgQuadClearDofCache(void **clist0, QUAD *quad, BOOLEAN final)
/* clear one or all cached DOF (basis) values in clist */
{
    QUAD_CACHE_LIST **clist;
    int i, n;

    if (quad == NULL)
	phgQuadGetFuncValues(NULL, NULL, 0, NULL, NULL);

    if (*(clist = (QUAD_CACHE_LIST **)clist0) == NULL)
	return;

    /* clear cache for 'quad' */
    if (quad != NULL) {
	if (quad->id == -1)
	    return;
	i = quad->id;
	assert(i < quad_list_count && quad_list[i] == quad);
	n = i + 1;
    }
    else {
	i = 0;
	n = (*clist)->n;
    }

    for (; i < n; i++) {
	if ((*clist)->caches[i] == NULL)
	    continue;
	(*clist)->caches[i]->e = NULL;
	phgFree((*clist)->caches[i]->data);
	(*clist)->caches[i]->data = NULL;
	if (final) {
	    phgFree((*clist)->caches[i]);
	    (*clist)->caches[i] = NULL;
	}
    }

    if (quad == NULL && final) {
	phgFree((*clist)->caches);
	phgFree(*clist);
	*clist = NULL;
    }
}

/*----------------- Basis and userfunc caching utilities -------------------*/

const FLOAT *
phgQuadGetFuncValues(GRID *g, SIMPLEX *e, int dim, DOF_USER_FUNC userfunc,
		     QUAD *quad)
/* cache and returns the function values of 'userfunc' at quadrature points of
 * 'quad'. if quad == NULL then clear cached values */
{
    static FLOAT *funcvalues = NULL;
    static QUAD *cached_quad = NULL;
    static SIMPLEX *cached_e = NULL;
    static INT cached_serial_no = -1;
    static DOF_USER_FUNC cached_userfunc = NULL;

    int i;
    FLOAT *v;
    const FLOAT *lambda;
    FLOAT x, y, z;

    if (quad == NULL) {
	/* clear cache */
	phgFree(funcvalues);
	funcvalues = NULL;
	cached_quad = NULL;
	cached_e = NULL;
	cached_serial_no = -1;
	cached_userfunc = NULL;
	return NULL;
    }

    if (cached_quad != quad || cached_e != e || cached_userfunc != userfunc
	|| cached_serial_no != g->serial_no) {
	cached_quad = quad;
	cached_e = e;
	cached_serial_no = g->serial_no;
	cached_userfunc = userfunc;

	/* compute function values */
	lambda = quad->points;
	phgFree(funcvalues);
	funcvalues = phgAlloc(dim * quad->npoints * sizeof(*funcvalues));
	for (i = 0, v = funcvalues; i < quad->npoints; i++, v += dim) {
	    phgGeomLambda2XYZ(g, e, lambda, &x, &y, &z);
	    userfunc(x, y, z, v);
	    lambda += Dim + 1;
	}
    }

    return funcvalues;
}

const FLOAT *
phgQuadGetBasisValues(SIMPLEX *e, DOF *u, int n, QUAD *quad)
/* returns the address of an array containing the values of the n-th
 * basis function at the quadrature points of 'quad'. The values are
 * returned in the order FLOAT[quad->npoints][type->dim]
 *
 * Note: in the internal buffer the values are stored as:
 *	FLOAT[nbas][quad->npoints][type->dim]
 */
{
    DOF_TYPE *type;
    QUAD_CACHE *cache;
    const FLOAT *v, *lambda;
    FLOAT *bas;
    int i, j, nbas;

    assert(!SpecialDofType(u->type));

    if (!DofIsHP(u)) {
	type = u->type;
	nbas = type->nbas;
	cache = get_cache(&type->cache_basfunc, quad);
    }
    else {
	type = u->hp->info->types[u->hp->max_order];
	nbas = DofGetNBas(u, e);
	cache = get_cache(&u->hp->cache_basfunc, quad);
    }

    if (cache->data != NULL &&
	((!DofIsHP(u) && type->invariant) || cache->e == e))
	return cache->data + n * quad->npoints * type->dim;

    if (type->BasFuncs == NULL)
	phgError(1, "no basis function for DOF \"%s\"!\n", u->name);

    cache->e = e;
    phgFree(cache->data);
    bas = cache->data =
		phgAlloc(nbas * quad->npoints * type->dim * sizeof(*bas));
    lambda = quad->points;
    for (i = 0; i < quad->npoints; i++) {
	v = type->BasFuncs(u, e, 0, -1, lambda);
	bas = cache->data + i * type->dim;
	for (j = 0; j < nbas; j++) {
	    memcpy(bas, v, type->dim * sizeof(*v));
	    v += type->dim;
	    bas += quad->npoints * type->dim;
	}
	lambda += Dim + 1;
    }

    return cache->data + n * quad->npoints * type->dim;
}

static inline const FLOAT *
get_grad_lambda(SIMPLEX *e, DOF *u, int n, QUAD *quad)
/* returns the address of an array containing the values of the lambda
 * gradient of the n-th basis function at the quadrature points of 'quad'.
 * The values are returned as:
 *	FLOAT[quad->npoints][type->dim][Dim + 1]
 *
 * Note: in the internal buffer the values are stored as:
 *	FLOAT[nbas][quad->npoints][type->dim][Dim + 1]
 */
{
    DOF_TYPE *type;
    QUAD_CACHE *cache;
    const FLOAT *v;
    FLOAT *grad, *lambda;
    int i, j, nbas, ngrad, nvalues;

    assert(!SpecialDofType(u->type));

    if (!DofIsHP(u)) {
	type = u->type;
	nbas = type->nbas;
	cache = get_cache(&u->type->cache_basgrad, quad);
    }
    else {
	type = u->hp->info->types[u->hp->max_order];
	nbas = DofGetNBas(u, e);
	cache = get_cache(&u->hp->cache_basgrad, quad);
    }
    ngrad = nbas * type->dim;

    if (cache->data != NULL &&
	((!DofIsHP(u) && type->invariant) || cache->e == e))
	return cache->data + n * quad->npoints * type->dim * (Dim + 1);

    if (type->BasGrads == NULL)
	phgError(1, "no gradient function for DOF \"%s\"!\n", u->name);

    cache->e = e;
    phgFree(cache->data);
    cache->data = phgAlloc(quad->npoints * ngrad * (Dim + 1) * sizeof(*grad));
    lambda = quad->points;
    nvalues = (Dim + 1) * type->dim;
    for (i = 0; i < quad->npoints; i++) {
	v = type->BasGrads(u, e, 0, -1, lambda);
	grad = cache->data + i * nvalues;
	for (j = 0; j < nbas; j++) {
	    memcpy(grad, v, nvalues * sizeof(*v));
	    v += nvalues;
	    grad += quad->npoints * nvalues;
	}
	lambda += Dim + 1;
    }

    return cache->data + n * quad->npoints * type->dim * (Dim + 1);
}

const FLOAT *
phgQuadGetBasisGradient(SIMPLEX *e, DOF *u, int n, QUAD *quad)
/* returns the xyz gradients of the n-th basis function on element e
 * The values are returned as
 *	FLOAT[quad->npoints][type->dim][Dim]
 *
 * Note: in the internal buffer the values are stored as:
 *	FLOAT[nbas][quad->npoints][type->dim][Dim]
 */
{
    DOF_TYPE *type;
    QUAD_CACHE *cache;
    int i, j, k;
    int nbas, nvalues, npoints = quad->npoints;
    FLOAT *data, *p;
    const FLOAT *J, *J1, *G;

    assert(!SpecialDofType(u->type));

    if (!DofIsHP(u)) {
	type = u->type;
	nbas = type->nbas;
	cache = get_cache(&u->type->cache_gradient, quad);
    }
    else {
	type = u->hp->info->types[u->hp->max_order];
	nbas = DofGetNBas(u, e);
	cache = get_cache(&u->hp->cache_gradient, quad);
    }
    nvalues = Dim * type->dim;

    if (cache->data != NULL && cache->e == e)
	return cache->data + n * npoints * nvalues;

    cache->e = e;
    phgFree(cache->data);
    data = cache->data = phgAlloc(nbas * npoints * nvalues * sizeof(*data));

    J = phgGeomGetJacobian(u->g, e);
    for (k = 0; k < nbas; k++, data += npoints * nvalues) {
	if (k < NVert && !DofIsHP(u) && type->name[0] == 'P'
	    && type->name[1] >= '0' && type->name[1] <= '9') {
	    /* faster code for special case */
	    p = data;
	    if (type == DOF_P0) {
		npoints *= Dim;
		for (i = 0; i < npoints; i++) {
		    *(p++) = 0.;
		}
	    }
	    else if (type == DOF_P1) {
		/* D\phi / D\lambda is the unit matrix */
		J1 = J + k * (Dim + 1);
		for (i = 0; i < npoints; i++) {
		    memcpy(p, J1, Dim * sizeof(*data));
		    p += Dim;
		}
	    }
	    else {
		/* the first NVert rows of D\phi / D\lambda are diagonal */
		G = get_grad_lambda(e, u, k, quad) + k;
		J1 = J + k * (Dim + 1);
		for (i = 0; i < npoints; i++) {
		    FLOAT d = *G;
		    *(p++) = d * J1[0];
		    *(p++) = d * J1[1];
		    *(p++) = d * J1[2];
		    G += Dim + 1;
		}
	    }
	}
	else {
	    G = get_grad_lambda(e, u, k, quad);
	    /* compute data[npoints][Dim] = G[npoints][Dim+1] * J[Dim+1][Dim] */
#if 0
	    if (npoints > 4) {
		static FLOAT a = 1.0, b = 0.0;
		int k = (Dim + 1) * type->dim, j = Dim * type->dim;
		F77name(dgemm) ("t", "t", &npoints, &j, &k, &a, G, &k, J, &k,
				&b, data, &k);
		return data;
	    }
#else
	    p = data;
	    for (i = 0; i < npoints; i++) {
		for (j = 0; j < type->dim; j++) {
		    *(p++) = G[0] * J[0 * (Dim + 1) + 0] +
			G[1] * J[1 * (Dim + 1) + 0] +
			G[2] * J[2 * (Dim + 1) + 0] +
			G[3] * J[3 * (Dim + 1) + 0];
		    *(p++) = G[0] * J[0 * (Dim + 1) + 1] +
			G[1] * J[1 * (Dim + 1) + 1] +
			G[2] * J[2 * (Dim + 1) + 1] +
			G[3] * J[3 * (Dim + 1) + 1];
		    *(p++) = G[0] * J[0 * (Dim + 1) + 2] +
			G[1] * J[1 * (Dim + 1) + 2] +
			G[2] * J[2 * (Dim + 1) + 2] +
			G[3] * J[3 * (Dim + 1) + 2];
		    G += Dim + 1;
		}
	    }
#endif
	}
    }

    return cache->data + n * npoints * nvalues;
}

const FLOAT *
phgQuadGetBasisCurl(SIMPLEX *e, DOF *u, int n, QUAD *quad)
/* returns the curl of the n-th basis functions at quadrature points 
 * the values are cached as FLOAT[nbas][quad->npoints][type->dim]
 * The requested values are returned as
 *	FLOAT[quad->npoints][type->dim]
 * (type->dim % Dim == 0)
 */
{
    DOF_TYPE *type;
    QUAD_CACHE *cache;
    FLOAT *curl;
    int i, j, nbas;
    int I, J;

    assert(!SpecialDofType(u->type));

    if (!DofIsHP(u)) {
	type = u->type;
	nbas = type->nbas;
	cache = get_cache(&u->type->cache_curl, quad);
    }
    else {
	type = u->hp->info->types[u->hp->max_order];
	nbas = DofGetNBas(u, e);
	cache = get_cache(&u->hp->cache_curl, quad);
    }

    if (type->dim % Dim != 0)
	phgError(1, "%s:%d, cannot compute curl of \"%s\" whose dimension is "
		 "not multiple of %d\n", u->name, Dim);

    if (cache->data != NULL && cache->e == e)
	return cache->data + n * quad->npoints * type->dim;

    cache->e = e;
    phgFree(cache->data);
    cache->data =
	phgAlloc(nbas * quad->npoints * type->dim * sizeof(*curl));

    if (!DofIsHP(u) && type == DOF_ND1) {
	/* faster code (?) for Nedelec-1 element */
	FLOAT (*nabla)[Dim + 1] = (void *)(phgGeomGetJacobian(u->g, e));
	for (i = 0; i < quad->npoints; i++) {
	    curl = cache->data + i * type->dim;
	    for (j = 0; j < nbas; j++) {
		GetEdgeVertices(e, j, I, J);
		curl[0] =
		    (nabla[I][1] * nabla[J][2] -
		     nabla[I][2] * nabla[J][1]) * 2.;
		curl[1] =
		    (nabla[I][2] * nabla[J][0] -
		     nabla[I][0] * nabla[J][2]) * 2.;
		curl[2] =
		    (nabla[I][0] * nabla[J][1] -
		     nabla[I][1] * nabla[J][0]) * 2.;
		curl += quad->npoints * type->dim;
	    }
	}
    }
    else {
	FLOAT (*g)[Dim];
	for (i = 0; i < quad->npoints; i++) {
	    curl = cache->data + i * type->dim;
	    for (j = 0; j < nbas; j++) {
		g = (void *)(phgQuadGetBasisGradient(e, u, j, quad) +
			     i * type->dim * Dim);
		curl[0] = g[2][1] - g[1][2];
		curl[1] = g[0][2] - g[2][0];
		curl[2] = g[1][0] - g[0][1];
		curl += quad->npoints * type->dim;
	    }
	}
    }

    return cache->data + n * quad->npoints * type->dim;
}

const FLOAT *
phgQuadGetDofValues(SIMPLEX *e, DOF *u, QUAD *quad)
/* returns the address of an array containing the values of dof
 * at the quadrature points of 'quad', the values are stored in the order
 * FLOAT[quad->npoints][DofDim(u)].
 */
{
    QUAD_CACHE *cache;
    FLOAT *v;
    const FLOAT *bas, *lambda;
    int nvalues = DofDim(u);
    int i;

    cache = get_cache(&u->cache_func, quad);
    if (cache->data != NULL && (u->invariant || cache->e == e))
	return cache->data;

    cache->e = e;
    phgFree(cache->data);
    v = cache->data = phgAlloc(quad->npoints * nvalues * sizeof(*v));
    /* Note: calling phgQuadGetBasisValues may further save recomputation
     * of basis functions if u->type->invariant == TRUE (or when computing
     * u dot u). By requesting values for basis function 0 we eventually
     * get the start address of values of all basis functions at the
     * quadrature points, stored as
     *          FLOAT[nbas][quad->npoints][type->dim]
     */
    if (!SpecialDofType(u->type) && !DofIsHP(u))
	bas = phgQuadGetBasisValues(e, u, 0, quad);
    else
	bas = NULL;
    lambda = quad->points;
    for (i = 0; i < quad->npoints; i++) {
	phgDofEval_(u, e, lambda, bas, quad->npoints * DofTypeDim(u), v, NULL);
	v += nvalues;
	if (bas != NULL)
	    bas += u->type->dim;
	lambda += Dim + 1;
    }
    return cache->data;
}

/*------------------------------ 2D functions --------------------------------*/

FLOAT
phgQuadFaceDofDotBas(SIMPLEX *e, int face, DOF *u, DOF_PROJ proj,
		     DOF *v, int N, int order)
{
    GRID *g = u->g;
    int i, j, k, nvalues, v0, v1, v2, dim;
    FLOAT d, d0, lambda[Dim + 1];
    FLOAT *dof, *buffer, *p0, *p1;
    const FLOAT *bas, *p, *w;
    QUAD *quad;
    const FLOAT *n = NULL;	/* out normal vector */
    DOF_TYPE *type;

    assert(!SpecialDofType(v->type));
    assert(face >= 0 && face <= 3);

    type = (DofIsHP(v) ? v->hp->info->types[v->hp->max_order] : v->type);

    if (order < 0) {
	i = DofTypeOrder(u, e);
	j = DofTypeOrder(v, e);
	if (i < 0)
	    i = j;
	order = i + j;
    }
    quad = phgQuadGetQuad2D(order);

    v0 = GetFaceVertex(face, 0);
    v1 = GetFaceVertex(face, 1);
    v2 = GetFaceVertex(face, 2);
    lambda[face] = 0.;

    dim = DofDim(u) / (proj == DOF_PROJ_DOT ? Dim : 1);
    nvalues = DofTypeDim(v);

    if (nvalues != dim)
	phgError(1, "%s: dimensions mismatch\n", __func__);

    buffer = phgAlloc(nvalues * sizeof(*buffer));
    p = quad->points;
    w = quad->weights;

    d = 0.;
    switch (proj) {
        case DOF_PROJ_NONE:
             for (i = 0; i < quad->npoints; i++) {
                 lambda[v0] = *(p++);
                 lambda[v1] = *(p++);
                 lambda[v2] = *(p++);
             
		 dof = phgDofEval(u, e, lambda, buffer);
                 bas = (FLOAT *)type->BasFuncs(v, e, N, N + 1, lambda);
                 d0 = 0.;
                 for (j = 0; j < nvalues; j++) {
                     d0 += *(bas++) * *(dof++);
		 }
		 d += d0 * *(w++);
	     }
	     
	     phgFree(buffer);
             break;
	     
        case DOF_PROJ_DOT:
             p0 = phgAlloc(nvalues * Dim * sizeof(*p0));
	     n = phgGeomGetFaceOutNormal(g, e, face);
             for (i = 0; i < quad->npoints; i++) {
                 lambda[v0] = *(p++);
                 lambda[v1] = *(p++);
                 lambda[v2] = *(p++);
                 
		 p1 = phgDofEval(u, e, lambda, p0);
                 bas = type->BasFuncs(v, e, N, N + 1, lambda);
		 dof = buffer;
		 for (k = 0; k < nvalues; k++) {
                      *(dof++) = n[0] * p1[0] + n[1] * p1[1] + n[2] * p1[2];
		      p1 += Dim;
		 }
                 d0 = 0.;
		 dof = buffer;
                 for (j = 0; j < nvalues; j++) {
                     d0 += *(bas++) * *(dof++);
		 }
		 d += d0 * *(w++);
	     }
	     
             phgFree(buffer);
             phgFree(p0);
             break;
	     
        case DOF_PROJ_CROSS:
             p0 = phgAlloc(nvalues * sizeof(*p0));
	     n = phgGeomGetFaceOutNormal(g, e, face);
             for (i = 0; i < quad->npoints; i++) {
                 lambda[v0] = *(p++);
                 lambda[v1] = *(p++);
                 lambda[v2] = *(p++);
                 
		 p1 = phgDofEval(u, e, lambda, p0);
                 bas = type->BasFuncs(v, e, N, N + 1, lambda);
		 dof = buffer;
		 for (k = 0; k < nvalues / Dim; k++) {
                      *(dof++) = p1[1] * n[2] - p1[2] * n[1];
		      *(dof++) = p1[2] * n[0] - p1[0] * n[2];
                      *(dof++) = p1[0] * n[1] - p1[1] * n[0];
		      p1 += Dim;
		 }
                 d0 = 0.;
		 dof = buffer;
                 for (j = 0; j < nvalues; j++) {
                     d0 += *(bas++) * *(dof++);
		 }
		 d += d0 * *(w++);
	     }
	     
             phgFree(buffer);
             phgFree(p0);
             break;
	     
        default:
             phgError(1, "%s: unknown projection %d\n", __func__, proj);
    }

    return d * phgGeomGetFaceArea(u->g, e, face);
}

FLOAT
phgQuadFaceDofDotDof(SIMPLEX *e, int face, DOF *u, DOF_PROJ proj,
		     DOF *v, int order)
{
    GRID *g = u->g;
    int i, j, k, nvalues, v0, v1, v2, dim;
    FLOAT d, d0, lambda[Dim + 1];
    FLOAT *dof, *buffer, *p0, *p1;
    FLOAT *dofv, *pv;
    const FLOAT *p, *w;
    QUAD *quad;
    const FLOAT *n = NULL;	/* out normal vector */

    assert(face >= 0 && face <= 3);

    if (order < 0) {
	i = DofTypeOrder(u, e);
	j = DofTypeOrder(v, e);
	if (i < 0)
	    i = j;
	order = i + j;
    }
    quad = phgQuadGetQuad2D(order);

    v0 = GetFaceVertex(face, 0);
    v1 = GetFaceVertex(face, 1);
    v2 = GetFaceVertex(face, 2);
    lambda[face] = 0.;

    dim = DofDim(u) / (proj == DOF_PROJ_DOT ? Dim : 1);
    nvalues = DofDim(v);

    if (nvalues != dim)
	phgError(1, "%s: dimensions mismatch\n", __func__);

    buffer = phgAlloc(nvalues * sizeof(*buffer));
    dofv = phgAlloc(nvalues * sizeof(*dofv));
    p = quad->points;
    w = quad->weights;

    d = 0.;
    switch (proj) {
        case DOF_PROJ_NONE:
             for (i = 0; i < quad->npoints; i++) {
                 lambda[v0] = *(p++);
                 lambda[v1] = *(p++);
                 lambda[v2] = *(p++);
             
		 dof = phgDofEval(u, e, lambda, buffer);
                 pv = phgDofEval(v, e, lambda, dofv);
                 d0 = 0.;
                 for (j = 0; j < nvalues; j++) {
                     d0 += *(pv++) * *(dof++);
		 }
		 d += d0 * *(w++);
	     }
	     
             phgFree(buffer);
	     phgFree(dofv);
             break;
	     
        case DOF_PROJ_DOT:
             p0 = phgAlloc(nvalues * Dim * sizeof(*p0));
	     n = phgGeomGetFaceOutNormal(g, e, face);
	     
             for (i = 0; i < quad->npoints; i++) {
                 lambda[v0] = *(p++);
                 lambda[v1] = *(p++);
                 lambda[v2] = *(p++);
                 
		 p1 = phgDofEval(u, e, lambda, p0);
		 pv = phgDofEval(v, e, lambda, dofv);
		 dof = buffer;
		 for (k = 0; k < nvalues; k++) {
                      *(dof++) = n[0] * p1[0] + n[1] * p1[1] + n[2] * p1[2];
		      p1 += Dim;
		 }
		 
                 d0 = 0.;
		 dof = buffer;
                 for (j = 0; j < nvalues; j++) {
                     d0 += *(pv++) * *(dof++);
		 }
		 d += d0 * *(w++);
	     }
	     
             phgFree(buffer);
	     phgFree(dofv);
             phgFree(p0);
             break;
	     
        case DOF_PROJ_CROSS:
	     assert(DofDim(u) % Dim == 0);
             p0 = phgAlloc(nvalues * sizeof(*p0));
	     n = phgGeomGetFaceOutNormal(g, e, face);
             
	     for (i = 0; i < quad->npoints; i++) {
                 lambda[v0] = *(p++);
                 lambda[v1] = *(p++);
                 lambda[v2] = *(p++);
                 
		 p1 = phgDofEval(u, e, lambda, p0);
		 pv = phgDofEval(v, e, lambda, dofv);
		 dof = buffer;
		 for (k = 0; k < nvalues / Dim; k++) {
                      *(dof++) = p1[1] * n[2] - p1[2] * n[1];
		      *(dof++) = p1[2] * n[0] - p1[0] * n[2];
                      *(dof++) = p1[0] * n[1] - p1[1] * n[0];
		      p1 += Dim;
		 }
                 d0 = 0.;
		 dof = buffer;
                 for (j = 0; j < nvalues; j++) {
                     d0 += *(pv++) * *(dof++);
		 }
		 d += d0 * *(w++);
	     }
	     
             phgFree(buffer);
	     phgFree(dofv);
             phgFree(p0);
             break;
	     
        default:
             phgError(1, "%s: unknown projection %d\n", __func__, proj);
    }
    return d * phgGeomGetFaceArea(u->g, e, face);
}

/* a x b */
#define OP_CROSS0(a,b)  \
	     ((*((a)+1)) * (*((b)+2)) - (*((a)+2)) * (*((b)+1)))
#define OP_CROSS1(a,b)  \
	     ((*((a)+2)) * (*(b)) - (*(a)) * (*((b)+2)))
#define OP_CROSS2(a,b)  \
	     ((*(a)) * (*((b)+1)) - (*((a)+1)) * (*(b)))

/* a \dot b */
#define OP_DOT(a,b) \
    ((*(a)) * (*(b)) + (*((a)+1)) * (*((b)+1)) + (*((a)+2)) * (*((b)+2)))

FLOAT *
phgQuadFaceADofCrossDof(SIMPLEX *e, int face, DOF *A, DOF *u, DOF_PROJ u_proj,
			DOF *v, DOF_PROJ v_proj, int order, FLOAT *reval)
/* \int A ( u \corss n ) \cross ( v \cross n )  */
{
    GRID *g = u->g;
    int i, j, v0, v1, v2;
    int udim, vdim, dim;
    FLOAT *pd, lambda[Dim + 1], alpha, area;
    FLOAT *dofu, *pu;
    FLOAT *dofv, *pv;
    FLOAT *buffer0, *buffer1, *p0 = NULL, *p1 = NULL;
    const FLOAT *p, *w;
    QUAD *quad;
    const FLOAT *n = NULL;	/* out normal vector */

    assert(face >= 0 && face <= 3);
    assert(u != NULL && (DofDim(u) % Dim == 0));
    assert(reval!=NULL);
    if(A!=NULL)
       assert(DofDim(A)==1);

    if (order < 0) {
	i = DofTypeOrder(u, e);
	if (v != NULL)
	    j = DofTypeOrder(v, e);
	else
	    j = 0;
	if (i < 0)
	    i = j;
	order = i + j;
    }
    quad = phgQuadGetQuad2D(order);

    v0 = GetFaceVertex(face, 0);
    v1 = GetFaceVertex(face, 1);
    v2 = GetFaceVertex(face, 2);
    lambda[face] = 0.;

    udim = DofDim(u) / (u_proj == DOF_PROJ_DOT ? Dim : 1);

    if (v != NULL){
	vdim = DofDim(v) / (v_proj == DOF_PROJ_DOT ? Dim : 1);
	if (vdim != udim)
	    phgError(1, "%s: dimensions mismatch\n", __func__);
    }

    dim = DofDim(u);
    dofu = phgAlloc(dim * sizeof(*dofu) * 4);
    dofv = dofu + dim;
    buffer0 = dofu + dim * 2;
    buffer1 = dofu + dim * 3;

    p = quad->points;
    w = quad->weights;

    pd=reval;
    for(i = 0; i < udim; i++) 
        (*pd++) = 0.;

    n = phgGeomGetFaceOutNormal(g, e, face);
    area = phgGeomGetFaceArea(g, e, face);
    for (i = 0; i < quad->npoints; i++, w++){
        lambda[v0] = *(p++);
        lambda[v1] = *(p++);
        lambda[v2] = *(p++);
        switch (u_proj) {
             /* u */
	    case DOF_PROJ_NONE:
		 p0 = phgDofEval(u, e, lambda, buffer0);
	         break;
	     /* ( (u \dot n) */
	    case DOF_PROJ_DOT:
		 pu = phgDofEval(u, e, lambda, dofu);
		 p0 = buffer0;
	         for (j = 0; j < dim/Dim ; j++) {
	             (*p0++) = OP_DOT (pu,n);
		     pu += Dim;
	  	 } 
	        break;
	     /* ( (u  \cross  n) */
            case DOF_PROJ_CROSS:
		 pu = phgDofEval(u, e, lambda, dofu);
		 p0 = buffer0;
                 for (j = 0; j < dim/Dim; j++) {
		     (*p0++) = OP_CROSS0(pu, n);
		     (*p0++) = OP_CROSS1(pu, n);
		     (*p0++) = OP_CROSS2(pu, n);
		     pu += Dim;
	         }
		 break;
           default:
                 phgError(1, "%s: unknown projection %d\n", __func__, u_proj);
        }
	
        if(v == NULL){
            p0=buffer0;
            pd=reval;
	    if (A != NULL){
	        phgDofEval(A, e, lambda, &alpha);
	        for (j = 0; j < udim ; j++)
	            (*pd++) += (*p0++)*(*w)*alpha; 
	    }else{
	        for(j = 0; j < udim ; j++)
	        (*pd++) += (*p0++)*(*w); 
	    }
        }else{
            switch(v_proj) {
              /* v */
	      case DOF_PROJ_NONE:
	           p1 = phgDofEval(v, e, lambda, buffer1);
	           break;
	      /* ( (v \dot n) */
	      case DOF_PROJ_DOT:
		   pv = phgDofEval(v, e, lambda, dofv);
		   p1 = buffer1;
	           for (j = 0; j < dim/Dim ; j++) {
	               (*p1++) = OP_DOT ( pv , n ) ;
		        pv += Dim;
	  	   } 
	           break;
	      /* ( (v  \cross  n) */
              case DOF_PROJ_CROSS:
		   pv = phgDofEval(v, e, lambda, dofv);
		   p1 = buffer1;
                   for (j = 0; j < dim/Dim; j++) {
		       (*p1++) = OP_CROSS0(pv, n);
		       (*p1++) = OP_CROSS1(pv, n);
		       (*p1++) = OP_CROSS2(pv, n);
		       pv += Dim;
	           }
		   break;
              default:
                   phgError(1, "%s: unknown projection %d\n", __func__, v_proj);
            }

            p0=buffer0;
            p1=buffer1;
            pd=reval;
            if (A!=NULL){
                phgDofEval(A, e, lambda, &alpha);
                for (j = 0; j < udim/Dim; j++){
	            (*pd++) += OP_CROSS0(p0, p1)*(*w)*alpha;
	            (*pd++) += OP_CROSS1(p0, p1)*(*w)*alpha;
	            (*pd++) += OP_CROSS2(p0, p1)*(*w)*alpha;
                    p0 += Dim;
	            p1 += Dim;
                }
            }else{
                for (j = 0; j < udim/Dim; j++) {
	            (*pd++) += OP_CROSS0(p0, p1)*(*w);
	            (*pd++) += OP_CROSS1(p0, p1)*(*w);
	            (*pd++) += OP_CROSS2(p0, p1)*(*w);
                    p0 += Dim;
	            p1 += Dim;
                }
           }
        }
    }/* end-for-i */ 

    for(i = 0; i< udim; i++, pd++)
        reval[i] *= area;
    phgFree(dofu);

    return reval;
}

FLOAT
phgQuadDofNormP(SIMPLEX *e, DOF *u, int order, int p)
{
    int i, j, nvalues;
    FLOAT d, d0, tmp;
    const FLOAT *v1, *w;
    QUAD *quad;

    assert(u != NULL);
    nvalues = DofDim(u);

    if (order < 0) {
        i = DofTypeOrder(u, e) * p;
        if (i < 0)
            i = 1;
	order = i;
    }
    quad = phgQuadGetQuad3D(order);

    d = 0.;
    w = quad->weights;
    v1 = phgQuadGetDofValues(e, u, quad);
    for (i = 0; i < quad->npoints; i++){
        tmp = 0.;
        for (j = 0; j < nvalues; j++){
            d0 = *(v1++);
            tmp += Pow(Fabs(d0), p);
	}
        d += tmp * *(w++);
    }

    return d * phgGeomGetVolume(u->g, e);
}

#if 0	/*------------------------ old implementation -----------------------*/

DOF *
phgQuadFaceJumpN(DOF *u, DOF_PROJ proj, const char *name, int order, DOF *gn)
/* returns a DOF with one value on each face which is the (L^2 norm)^2 over
 * the face of the jump of 'u' acrossing the face.
 *
 * 'proj' specifies the projection to use: DOF_PROJ_NONE means no projection,
 * DOF_PROJ_DOT means \f$\vec{u} \cdot \vec{n}\f$, DOF_PROJ_CROSS means
 * \f$\vec{u} \times \vec{n}\f$.
 *
 * 'name' specifies the name of the returned DOF.
 *
 * 'order' specifies numerical integration order, if 'order' < 0 then the
 * numerical integration order is set to 2 times the order of 'u'.
 * 
 * 'gn', if not NULL, specifies Neumann boundary condition. On a Neumann
 * face the jump is replaced by proj(u) - gn */
{
    static DOF_TYPE DOF_JUMP_ = { DofCache,
	"Face jump", NULL, NULL, NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, FE_None,
	TRUE, FALSE, -1, NFace, 0, -1, 1,
	0, 0, 1, 0
    };
    char s[1024];
    GRID *g = u->g;
    SIMPLEX *e, *e1;
    int i, j, k, dim, sign = 0;
    INT I;
    DOF *jump, *tmp;
    FLOAT values[DofDim(u)];
    FLOAT lambda[Dim + 1];
    FLOAT d, d0, *p0, *p1, *p2;
    FLOAT *w, *pts;
    const FLOAT *n = NULL;
#if USE_MPI
    NEIGHBOUR_DATA *neigh_data;
#endif
    int v0, v1, v2;
    QUAD *quad;

    /* first, compute jumps at face quadrature points */
    assert(!SpecialDofType(u->type));

    if (proj != DOF_PROJ_NONE && DofDim(u) % Dim != 0)
	phgError(1, "phgDofJumpFace: DOF \"%s\" has incorrect dimension for "
		 "the operation %d\n", u->name, proj);
    if (name == NULL) {
	sprintf(s, "Jumps of %s", u->name);
	name = s;
    }

    /* compute and store jumps at all quadrature points.
     *
     * FIXME: may need large amount of memory if order is high, better
     *	      to use phgDofInitNeighbourData to get remote DOF on shared
     *	      faces then integrate face by face, assuming the values on a
     *	      face only depends on DOF on the face */
    if (order < 0) {
#warning FIXME!
	order = (!DofIsHP(u) ? u->type->order : 2) * 2;
    }
    quad = phgQuadGetQuad2D(order);
    dim = DofDim(u) / (proj == DOF_PROJ_DOT ? Dim : 1);
    if (gn != NULL)
        assert(DofDim(gn) == dim);

    jump = phgDofNew(g, &DOF_JUMP_, dim * quad->npoints, "Jump tmp",
			DofNoAction);
    ForAllElements(g, e) {
	BOOLEAN flag;
	for (i = 0; i < NFace; i++) {
	    /* determine direction of the face normal */
	    if (HasLocalNeighbour(e, i)) {
		/* towards the vertex with larger index */
		e1 = e->neighbours[i];
		sign = e1->verts[phgOppositeVertex(g, e, i, e1)] - e->verts[i];
		flag = TRUE;
	    }
#if USE_MPI
	    else if (HasRemoteNeighbour(e, i)) {
		/* towards the submesh with larger rank */
		sign = GetRNeighbour(g, e, i)->rank - g->rank;
		flag = TRUE;
	    }
#endif
	    else {
		/* boundary face, do nothing (FIXME: set to 0?) */
		flag = FALSE;
	    }

	    if (flag) {
	        if (proj != DOF_PROJ_NONE) {
		    n = phgGeomGetFaceNormal(g, e, i);
		    if (sign < 0) {
		        static FLOAT m[Dim];
		        m[0] = -n[0];
		        m[1] = -n[1];
		        m[2] = -n[2];
		        n = m;
		    }
	        }
	        pts = quad->points;
	        lambda[i] = 0.;
	        GetFaceVertices(e, i, v0, v1, v2);
	        p0 = DofFaceData(jump, e->faces[i]);
	        for (k = 0; k < quad->npoints; k++) {
		    lambda[v0] = *(pts++);
		    lambda[v1] = *(pts++);
		    lambda[v2] = *(pts++);
		    phgDofEval(u, e, lambda, values);
		    p1 = values;
		    switch (proj) {
		        case DOF_PROJ_NONE:
			    if (sign > 0) {
			        for (j = 0; j < dim; j++)
				    *(p0++) += *(p1++);
			    }
			    else {
			        for (j = 0; j < dim; j++)
				    *(p0++) -= *(p1++);
			    }
			    break;
		        case DOF_PROJ_DOT:
			    for (j = 0; j < dim; j++) {
			        *(p0++) +=
				    n[0] * p1[0] + n[1] * p1[1] + n[2] * p1[2];
			        p1 += Dim;
			    }
			    break;
		        case DOF_PROJ_CROSS:
			    for (j = 0; j < dim / Dim; j++) {
			        *(p0++) += p1[1] * n[2] - p1[2] * n[1];
			        *(p0++) += p1[2] * n[0] - p1[0] * n[2];
			        *(p0++) += p1[0] * n[1] - p1[1] * n[0];
			        p1 += Dim;
			    }
			    break;
		        default:
			    phgError(1, "%s:%d undefined projection.\n",
				 __FILE__, __LINE__);
		    }		/* endof switch */
	        }		/* end of k-loop */
	    }
            else { 
		if (gn != NULL && (e->bound_type[i] & NEUMANN)) {
		    FLOAT vg[DofDim(gn)];

		    if (proj != DOF_PROJ_NONE)
		        n = phgGeomGetFaceOutNormal(g, e, i);

		    pts = quad->points;
		    lambda[i] = 0.;
		    GetFaceVertices(e, i, v0, v1, v2);
		    p0 = DofFaceData(jump, e->faces[i]);
		    for (k = 0; k < quad->npoints; k++) {
		        lambda[v0] = *(pts++);
		        lambda[v1] = *(pts++);
		        lambda[v2] = *(pts++);
		        phgDofEval(u, e, lambda, values);
		        phgDofEval(gn, e, lambda, vg);
		        p1 = values;
		        p2 = vg;

			switch (proj) {
			    case DOF_PROJ_NONE:
		 	         for (j = 0; j < dim; j++)
				     *(p0++) = *(p2++) - *(p1++);
				 break;
			    case DOF_PROJ_DOT:
			         for (j = 0; j < dim; j++) {
				     *(p0++) = *(p2++) - 
				         n[0]*p1[0] - n[1]*p1[1] - n[2]*p1[2];
				     p1 += Dim;
				 }
				 break;
			    case DOF_PROJ_CROSS:
				 for (j = 0; j < dim / Dim; j++) {
				     *(p0++) = *(p2++) - p1[1]*n[2]+p1[2]*n[1];
				     *(p0++) = *(p2++) - p1[2]*n[0]+p1[0]*n[2];
				     *(p0++) = *(p2++) - p1[0]*n[1]+p1[1]*n[0];
				     p1 += Dim;
				 }
				 break;
			    default:
				 phgError(1, "%s:%d undefined projection.\n",
					 __FILE__, __LINE__);
			}	/* endof switch */
		    }		/* end of k-loop */
		}		/* gn != NULL && boundary type is NEUMANN */
	    }			/* flag ifelse  */
	}			/* end of i-loop */
    }				/* end of ForAllElements */

#if USE_MPI
    neigh_data = phgDofInitNeighbourData(jump, NULL);
    for (I = 0; I < g->neighbours.count; I++) {
	RNEIGHBOUR *rn = g->neighbours.list + I;
	e = rn->local;
	p0 = DofFaceData(jump, e->faces[rn->vertex]);
	p1 = phgDofNeighbourData(neigh_data, e, rn->vertex, 0, NULL);
	for (k = 0; k < jump->type->np_face * jump->dim; k++)
	    *(p0++) += *(p1++);
    }
    phgDofReleaseNeighbourData(&neigh_data);
#endif

    /* then, integrate (sum) on each face */
    tmp = jump;
    jump = phgDofNew(g, &DOF_JUMP_, 1, name, DofNoAction);
    p0 = DofFaceData(tmp, 0);
    p1 = DofFaceData(jump, 0);
    if (dim > 1) {
	for (I = 0; I < g->nface; I++) {
	    d = 0.0;
	    w = quad->weights;
	    for (k = 0; k < quad->npoints; k++, w++) {
		d0 = 0.;
		for (i = 0; i < dim; i++, p0++)
		    d0 += (*p0) * (*p0);
		d += d0 * (*w);
	    }
	    *(p1++) = phgGeomGetFaceAreaByIndex(g, I) * d;
	}
    }
    else {
	for (I = 0; I < g->nface; I++) {
	    d = 0.0;
	    w = quad->weights;
	    for (k = 0; k < quad->npoints; k++, p0++, w++)
		d += (*p0) * (*p0) * (*w);
	    *(p1++) = phgGeomGetFaceAreaByIndex(g, I) * d;
	}
    }
    phgDofFree(&tmp);
    return jump;
}

#else	/*------------------------ new implementation -----------------------*/

DOF *
phgQuadFaceJumpN(DOF *u, DOF_PROJ proj, const char *name, int order, DOF *gn)
/* returns a DOF with one value on each face which is the (L^2 norm)^2 over
 * the face of the jump of 'u' acrossing the face.
 *
 * 'proj' specifies the projection to use: DOF_PROJ_NONE means no projection,
 * DOF_PROJ_DOT means \f$\vec{u} \cdot \vec{n}\f$, DOF_PROJ_CROSS means
 * \f$\vec{u} \times \vec{n}\f$.
 *
 * 'name' specifies the name of the returned DOF.
 *
 * 'order' specifies numerical integration order, if 'order' < 0 then the
 * numerical integration order is set to 2 times the order of 'u'.
 * 
 * 'gn', if not NULL, specifies Neumann boundary condition, on a Neumann
 * face the jump is replaced by:
 * 	 u - gn		if 'proj' == DOF_PROJ_NONE
 * 	 u.n - gn	if 'proj' == DOF_PROJ_DOT, FIXME: use (u-gn).n instead?
 * 	 uxn - gn	if 'proj' == DOF_PROJ_CROSS
 */
{
    static DOF_TYPE DOF_JUMP_ = { DofCache,
	"Face jump", NULL, NULL, NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, FE_None,
	TRUE, FALSE, -1, NFace, 0, -1, 1,
	0, 0, 1, 0
    };
    char s[1024];
    GRID *g = u->g;
    SIMPLEX *e;
    int ii, jj, face, j, k, dim, dim0 = DofDim(u);
    INT i;
    DOF *jump, *tmp = NULL;
    DOF_TYPE *type;
    FLOAT values[dim0], values1[dim0];
    FLOAT lambda[Dim + 1];
    FLOAT d, d0, d1, *p, *p0, *p1;
    FLOAT *w, *pts, *buffer = NULL;
    const FLOAT *n = NULL, *bas, *pb;
    NEIGHBOUR_DATA *neigh;
    SHORT *bases0 = NULL, *bases1 = NULL;
    int nbases0 = 0, nbases1 = 0, nbas0, nbas1, v0, v1, v2;
    QUAD *quad;
    VEF_MAP *vef;
    BOOLEAN neumann, cache_flag;
    FE_SPACE fe_space = (!DofIsHP(u) ? u->type->fe_space :
				       u->hp->info->fe_space);

    assert(!SpecialDofType(u->type));	/* return 0 in this case? */

    /* Note: only DG type is allowed (to be fixed when needed) */
    if (fe_space != FE_L2 && fe_space != FE_None)
	phgError(1, "%s: can't compute jumps of DOF \"%s\" "
		    "(only DG types are allowed).\n", __func__, u->name);

    if (proj != DOF_PROJ_NONE && DofDim(u) % Dim != 0)
	phgError(1, "phgDofJumpFace: DOF \"%s\" has incorrect dimension for "
		 "the operation %d\n", u->name, proj);

    dim = dim0 / (proj == DOF_PROJ_DOT ? Dim : 1);
    assert(gn == NULL || DofDim(gn) == dim);

    if (name == NULL) {
	sprintf(s, "Jumps of %s", u->name);
	name = s;
    }
    jump = phgDofNew(g, &DOF_JUMP_, 1, name, DofNoAction);

    vef = phgDofSetupVEFMap(g, jump, FACE_FLAG);
    neigh = phgDofInitNeighbourData(u, NULL);

    if (order < 0 && !DofIsHP(u))
	order = u->type->order * 2;

    if (order >= 0)
	quad = phgQuadGetQuad2D(order);
    else
	quad = NULL;		/* make gcc happy */

    if ((type = u->type) == NULL) {
	/* create a temperary non hp DOF whose type will be set by
	 * phgDofNeighbourNBas() to match the order of the neighbour.
	 * This DOF is used to get the list of the bases for the neighbouring
	 * element (bases1). */
	tmp = phgDofNew(g, type = u->hp->info->types[u->hp->max_order], u->dim,
			"temp", DofNoData);
	phgDofUnrefId(tmp->type);
    }

    cache_flag = FALSE;
    for (i = 0; i < g->nface; i++) {
	if (g->types_face[i] == UNREFERENCED)
	    continue;
	e = vef->Fmap[i];
	face = vef->Find[i];

	if (!(e->bound_type[face] & (INTERIOR | NEUMANN))) 
	    continue;		/* non Neumann boundary */

	neumann = (e->bound_type[face] & NEUMANN);
	if (neumann && gn == NULL)
	    continue;

	nbas0 = phgDofGetBasesOnFace(u, e, face, NULL);
	if (!neumann) {
	    if (!DofIsHP(u)) {
		nbas1 = nbas0;
	    }
	    else {
		nbas1 = phgDofNeighbourNBas(neigh, e, face, &tmp->type);
		if (order < 0) {
		    assert(DofIsHP(u));
		    j = u->hp->info->types[u->hp->elem_order[e->index]]->order;
		    quad = phgQuadGetQuad2D(j + tmp->type->order);
		}
	    }
	}
	else {
	    nbas1 = 0;
	    if (order < 0) {
		assert(DofIsHP(u));
		j = u->hp->info->types[u->hp->elem_order[e->index]]->order;
		if (gn->type == DOF_CONSTANT)
		    k = 0;
		else if (gn->type == DOF_ANALYTIC)
		    k = j;
		else
		    k = DofFaceOrder(gn, i);
		quad = phgQuadGetQuad2D(j + k);
	    }
	}

	if (nbases0 < nbas0) {
	    nbases0 = nbas0;
	    phgFree(bases0);
	    bases0 = phgAlloc(nbases0 * sizeof(*bases0));
	}
	phgDofGetBasesOnFace(u, e, face, bases0);

	if (!neumann && nbas0 != nbas1) {
	    if (nbases1 < nbas1) {
		nbases1 = nbas1;
		phgFree(bases1);
		bases1 = phgAlloc(nbases1 * sizeof(*bases1));
	    }
	    phgDofGetBasesOnFace(tmp, e, face, bases1);
	}

	GetFaceVertices(e, face, v0, v1, v2);

	pts = quad->points;
	w   = quad->weights;
	lambda[face] = 0.;
	d = 0.;
	for (k = 0; k < quad->npoints; k++) {
	    lambda[v0] = *(pts++);
	    lambda[v1] = *(pts++);
	    lambda[v2] = *(pts++);
	    memset(values, 0, dim0 * sizeof(*values));
	    for (j = 0; j < nbas0; j++) {
		ii = bases0[j];
		if (/*FALSE &&*/ !DofIsHP(u)) {
		    /* compute and cache bases values at quad points */
		    if (!cache_flag) {
			static int max, min;
			if (buffer == NULL) {
			    buffer = phgAlloc(quad->npoints * nbas0 * type->dim
						* sizeof(*buffer));
			    min = max = bases0[0];
			    for (jj = 1; jj < nbas0; jj++) {
				if (min > bases0[jj])
				    min = bases0[jj];
				else if (max < bases0[jj])
				    max = bases0[jj];
			    }
			} 
			p = buffer + k * nbas0 * type->dim;
			bas = type->BasFuncs(u, e, min, max + 1, lambda);
			for (jj = 0; jj < nbas0; jj++, p += type->dim)
			    memcpy(p, bas + (bases0[jj] - min) * type->dim,
						type->dim * sizeof(*buffer));
		    }
		    bas = buffer + (k * nbas0 + j) * type->dim;
		}
		else {
		    bas = type->BasFuncs(u, e, ii, ii + 1, lambda);
		}
		p0 = u->data + phgDofMapE2D(u, e, ii * u->dim);
		p = values;
		if (neumann || nbas0 != nbas1) {
		    for (ii = 0; ii < u->dim; ii++) {
			d0 = *(p0++);
			for (jj = 0, pb = bas; jj < type->dim; jj++)
			    *(p++) += d0 * *(pb++);
		    }
		}
		else {
		    p1 = phgDofNeighbourData(neigh, e, face, j, NULL);
		    for (ii = 0; ii < u->dim; ii++) {
			d0 = *(p0++) - *(p1++);
			for (jj = 0, pb = bas; jj < type->dim; jj++)
			    *(p++) += d0 * *(pb++);
		    }
		}
	    }
#if 0
/*if (i == 4)*/ {
FLOAT values0[dim0], lambda1[Dim + 1];
int u0, u1, u2;
phgDofEval(u, e, lambda, values0);
SIMPLEX *e1 = e->neighbours[face];
ii = phgOppositeVertex(g, e, face, e1);
GetFaceVertices(e1, ii, u0, u1, u2);
pts -= 3;
lambda1[ii] = 0.;
lambda1[u0] = *(pts++);
lambda1[u1] = *(pts++);
lambda1[u2] = *(pts++);
phgDofEval(u, e1, lambda1, values1);
for (ii = 0; ii < dim0; ii++)
phgInfo(-1, "face=%d, k=%d, ii = %d, value=%e, error=%e\n", i, k, ii, values[ii], values[ii] - values0[ii] + values1[ii]);
}
#endif

	    if (neumann) {
		phgDofEval(gn, e, lambda, values1);
	    }
	    else if (nbas0 != nbas1) {
		for (j = 0; j < nbas1; j++) {
		    ii = bases1[j];
		    bas = type->BasFuncs(tmp, e, ii, ii + 1, lambda);
		    p0 = phgDofNeighbourData(neigh, e, face, j, NULL);
		    p = values;
		    for (ii = 0; ii < u->dim; ii++) {
			d0 = *(p0++);
			for (jj = 0, pb = bas; jj < type->dim; jj++)
			    *(p++) -= d0 * *(pb++);
		    }
		}
	    }

	    /* projection */
	    d0 = 0.;
	    switch (proj) {
	        case DOF_PROJ_NONE:
		    if (neumann) {
			p1 = values1;
			for (ii = 0, p = values; ii < dim0; ii++) {
			    d1 = *(p++) - *(p1++);
			    d0 += d1 * d1;
			}
		    }
		    else {
			for (ii = 0, p = values; ii < dim0; ii++, p++)
			    d0 += *p * *p;
		    }
		    break;
	        case DOF_PROJ_DOT:
		    n = phgGeomGetFaceNormal(g, e, face);
		    if (neumann) {
			p1 = values1;
			for (ii = 0, p = values; ii < dim; ii++, p += Dim) {
			    d1 = n[0]*p[0] + n[1]*p[1] + n[2]*p[2] - *(p1++);
		            d0 += d1 * d1;
			}
		    }
		    else {
			for (ii = 0, p = values; ii < dim; ii++, p += Dim) {
			    d1 = n[0] * p[0] + n[1] * p[1] + n[2] * p[2];
		            d0 += d1 * d1;
			}
		    }
		    break;
	        case DOF_PROJ_CROSS:
		    n = phgGeomGetFaceNormal(g, e, face);
		    if (neumann) {
			p1 = values1;
			for (ii = 0, p = values; ii < dim / Dim; ii++) {
		            d1 = p[1] * n[2] - p[2] * n[1] - *(p1++);
			    d0 += d1 * d1;
		            d1 = p[2] * n[0] - p[0] * n[2] - *(p1++);
			    d0 += d1 * d1;
		            d1 = p[0] * n[1] - p[1] * n[0] - *(p1++);
			    d0 += d1 * d1;
			    p += Dim;
			}
		    }
		    else {
			for (ii = 0, p = values; ii < dim / Dim; ii++) {
		            d1 = p[1] * n[2] - p[2] * n[1];
			    d0 += d1 * d1;
		            d1 = p[2] * n[0] - p[0] * n[2];
			    d0 += d1 * d1;
		            d1 = p[0] * n[1] - p[1] * n[0];
			    d0 += d1 * d1;
			    p += Dim;
			}
		    }
		    break;
	        default:
		    phgError(1, "%s: invalid projection %d.\n", __func__, proj);
	    }		/* endswitch */
	    d += d0 * *(w++);
	}	/* endfor k */
	cache_flag = TRUE;
	*DofFaceData(jump, i) = d * phgGeomGetFaceAreaByIndex(g, i);
    }	/* endfor i */

    phgFree(buffer);
    phgFree(bases0);
    phgFree(bases1);
    if (tmp != NULL) {
	tmp->type = NULL;	/* shouldn't call phgDofUnrefId */
	phgDofFree(&tmp);
    }

    phgDofFreeVEFMap(&vef);
    phgDofReleaseNeighbourData(&neigh);

    return jump;
}

#endif	/*------------------------ new implementation -----------------------*/

/*------------------------------ 3D functions --------------------------------*/

FLOAT
phgQuadDofDotDof(SIMPLEX *e, DOF *u, DOF *v, int order)
/* computes \int u \cdot v on element 'e' (inner product),
 * using quadrature rule 'quad'.
 */
{
    int i, j, nvalues;
    FLOAT d, d0;
    const FLOAT *v1, *v2, *w;
    QUAD *quad;

    nvalues = DofDim(u);
    if (v != NULL && nvalues != DofDim(v))
	phgError(1, "%s:%d: dimensions of \"%s\" and \"%s\" don't match\n",
		 __FILE__, __LINE__, u->name, v->name);

    if (order < 0) {
	i = DofTypeOrder(u, e);
	j = DofTypeOrder(v, e);
	if (i < 0 && j < 0) {
	    phgInfo(-1, "phgQuadDofDotDof: don't use QUAD_DEFAULT when both "
		    "DOF types are analytic.\n");
	    phgError(1, "phgQuadDofDotDof: can't determine quadrature order, "
			"abort.\n");
	}
	if (i < 0)
	    i = j;
	else if (j < 0)
	    j = i;
	order = i + j;
    }
    quad = phgQuadGetQuad3D(order);

    w = quad->weights;
    v1 = phgQuadGetDofValues(e, u, quad);
    v2 = phgQuadGetDofValues(e, v, quad);

    if (nvalues == 1) {		/* faster code for special case */
	d = 0.;
	for (i = 0; i < quad->npoints; i++)
	    d += *(v1++) * *(v2++) * *(w++);
    }
    else {
	d = 0.;
	for (i = 0; i < quad->npoints; i++) {
	    d0 = 0.;
	    for (j = 0; j < nvalues; j++)
		d0 += *(v1++) * *(v2++);
	    d += d0 * *(w++);
	}
    }

    return d * phgGeomGetVolume(u->g, e);
}

FLOAT
phgQuadGradBasAGradBas(SIMPLEX *e, DOF *u, int n, DOF *A, DOF *v, int m,
		       int order)
{
    int i, j, k, nvalues = DofTypeDim(u), nn, An;
    FLOAT d, d0;
    const FLOAT *g1, *g2, *g2_tmp, *w, *coef;
    QUAD *quad;

    assert(!SpecialDofType(u->type) && !SpecialDofType(v->type));
    assert(nvalues == DofTypeDim(v));

    if (order < 0) {
	order = BasisOrder(u, e, n) - 1 + BasisOrder(v, e, m) - 1;
	if (A != NULL && DofTypeOrder(A, e) >= 0)
	    order += DofTypeOrder(A, e);
	if (order < 0)
	    order = 0;
    }
    quad = phgQuadGetQuad3D(order);

    d = 0.;
    g1 = phgQuadGetBasisGradient(e, u, n, quad);
    g2 = phgQuadGetBasisGradient(e, v, m, quad);
    w = quad->weights;
    if (NULL == A) {
	for (i = 0; i < quad->npoints; i++) {
	    d0 = 0.;
	    for (j = 0; j < nvalues; j++) {
		d0 += g1[0] * g2[0] + g1[1] * g2[1] + g1[2] * g2[2];
		g1 += Dim;
		g2 += Dim;
	    }
	    d += d0 * (*(w++));
	}
	return d * phgGeomGetVolume(u->g, e);
    }

    An = DofDim(A);
    coef = phgQuadGetDofValues(e, A, quad);
    if (1 == An) {
	for (i = 0; i < quad->npoints; i++) {
	    d0 = 0.;
	    for (j = 0; j < nvalues; j++) {
		d0 += g1[0] * g2[0] + g1[1] * g2[1] + g1[2] * g2[2];
		g1 += Dim;
		g2 += Dim;
	    }
	    d += d0 * (*(w++)) * *(coef++);
	}
    }
    else if (An == (nn = nvalues * Dim)) {
	for (i = 0; i < quad->npoints; i++) {
	    d0 = 0.;
	    for (j = 0; j < nvalues; j++) {
		d0 +=
		    g1[0] * g2[0] * coef[0] + g1[1] * g2[1] * coef[1] +
		    g1[2] * g2[2] * coef[2];
		g1 += Dim;
		g2 += Dim;
		coef += Dim;
	    }
	    d += d0 * (*(w++));
	}
    }
    else if (An == nn * nn) {
	for (i = 0; i < quad->npoints; i++) {
	    d0 = 0;
	    g2_tmp = g2;
	    for (j = 0; j < nn; j++) {
		g2 = g2_tmp;
		for (k = 0; k < nn; k++) {
		    d0 += *(g1) * *(g2++) * *(coef++);
		}
		g1++;
	    }
	    d += d0 * *(w++);
	}
    }
    else {
	phgError(1, "%s:%d: dimensions mismatch: coefficent A  %s (%d)",
		 __FILE__, __LINE__, DofTypeName(A), DofTypeDim(A));
    }

    return d * phgGeomGetVolume(u->g, e);
}

FLOAT *
phgQuadDofTimesBas(SIMPLEX *e, DOF *u, DOF *v, int n, int order, FLOAT *res)
/* computes \int 'u' * ('n'-th basis function of 'v') on element 'e'
 * using quadrature rule 'quad'. The results are returned in the user provided
 * array 'res' whose size should be greater than or equal to u->dim.
 */
{
    int i, j, nvalues;
    const FLOAT *bas, *w, *f;
    FLOAT d;
    QUAD *quad;

    assert(v != NULL && !SpecialDofType(v->type) && DofTypeDim(v) == 1);

    if (order < 0) {
	i = DofTypeOrder(u, e);
	j = DofTypeOrder(v, e);
	if (i < 0)
	    i = j;
	order = i + j;
    }
    quad = phgQuadGetQuad3D(order);

    nvalues = DofDim(u);
    f = phgQuadGetDofValues(e, u, quad);
    bas = phgQuadGetBasisValues(e, v, n, quad);

    w = quad->weights;
    if (nvalues == 1) {		/* faster code for special case */
	d = 0.0;
	for (i = 0; i < quad->npoints; i++) {
	    d += *(bas++) * *(w++) * *(f++);
	}
	res[0] = d * phgGeomGetVolume(u->g, e);
    }
    else {
	for (j = 0; j < nvalues; j++)
	    res[j] = 0;
	for (i = 0; i < quad->npoints; i++) {
	    d = *(bas++) * *(w++);
	    for (j = 0; j < nvalues; j++)
		res[j] += d * *(f++);
	}
	d = phgGeomGetVolume(u->g, e);
	for (j = 0; j < nvalues; j++)
	    res[j] *= d;
    }

    return res;
}

#if 0				/* note: use phgQuadBasABas */
/* computes integration of 
 *  	('n'-th basis function of 'u') 
 *  	\cdot 
 *  	('m'-th basis function of 'v')
 * on element 'e' (inner product), using quadrature rule 'quad'.
 */
FLOAT
phgQuadBasDotBas(SIMPLEX *e, DOF *u, int n, DOF *v, int m, int order)
{
    int i, j, nvalues = DofTypeDim(u);
    FLOAT *g1, *g2, *w;
    FLOAT d, d0;
    QUAD *quad;

    assert(!SpecialDofType(u->type) && !SpecialDofType(v->type));

    if (nvalues != DofTypeDim(v))
	phgError(1, "%s:%d: dimensions mismatch: %s (%d) <==> %s ($d))\n",
		 __FILE__, __LINE__, DofTypeName(u), DofTypeDim(u),
		 DofTypeName(v), DofTypeDim(v));

    if (order < 0)
	order = BasisOrder(u, e, n) + BasisOrder(v, e, m);
    quad = phgQuadGetQuad3D(order);

    g1 = phgQuadGetBasisValues(e, u, n, quad);
    g2 = phgQuadGetBasisValues(e, v, m, quad);
    d = 0.;
    w = quad->weights;
    for (i = 0; i < quad->npoints; i++) {
	d0 = 0.;
	for (j = 0; j < nvalues; j++) {
	    d0 += *(g1++) * (*(g2++));
	}
	d += d0 * (*(w++));
    }
    return d * phgGeomGetVolume(u->g, e);
}
#endif

FLOAT
phgQuadBasABas(SIMPLEX *e, DOF *u, int n, DOF *A, DOF *v, int m, int order)
/* computes integration of 
 *  	('n'-th basis function of 'u') 
 *  	\cdot 
 *       (coefficient function 'A')
 *       \cdot
 *  	('m'-th basis function of 'v')
 * on element 'e' (inner product), using quadrature rule 'quad'.
 *
 */
{
    int i, j, k, nvalues = DofTypeDim(u), An;
    const FLOAT *g1, *g2, *g2_tmp, *coef, *w;
    FLOAT d, d0;
    QUAD *quad;

    assert(!SpecialDofType(u->type) && !SpecialDofType(v->type));

    if (nvalues != DofTypeDim(v))
	phgError(1, "%s:%d: dimensions mismatch: %s (%d) <==> %s ($d))\n",
		 __FILE__, __LINE__, DofTypeName(u), DofTypeDim(u),
		 DofTypeName(v), DofTypeDim(v));

    if (order < 0) {
	order = BasisOrder(u, e, n) + BasisOrder(v, e, m);
	if (A != NULL && DofTypeOrder(A, e) >= 0)
	    order += DofTypeOrder(A, e);
    }
    quad = phgQuadGetQuad3D(order);

    g1 = phgQuadGetBasisValues(e, u, n, quad);
    g2 = phgQuadGetBasisValues(e, v, m, quad);
    d = 0.;
    w = quad->weights;
    if (NULL == A) {
	for (i = 0; i < quad->npoints; i++) {
	    d0 = 0.;
	    for (j = 0; j < nvalues; j++) {
		d0 += *(g1++) * *(g2++);
	    }
	    d += d0 * *(w++);
	}
	return d * phgGeomGetVolume(u->g, e);
    }

    An = DofDim(A);
    coef = phgQuadGetDofValues(e, A, quad);
    if (1 == An) {
	for (i = 0; i < quad->npoints; i++) {
	    d0 = 0.;
	    for (j = 0; j < nvalues; j++) {
		d0 += *(g1++) * (*(g2++));
	    }
	    d += d0 * (*(w++)) * *(coef++);
	}
    }
    else if (An == nvalues) {
	for (i = 0; i < quad->npoints; i++) {
	    d0 = 0.;
	    for (j = 0; j < nvalues; j++) {
		d0 += *(g1++) * *(g2++) * *(coef++);
	    }
	    d += d0 * *(w++);
	}
    }
    else if (An == nvalues * nvalues) {
	for (i = 0; i < quad->npoints; i++) {
	    d0 = 0.;
	    g2_tmp = g2;
	    for (j = 0; j < nvalues; j++) {
		g2 = g2_tmp;
		for (k = 0; k < nvalues; k++) {
		    d0 += *(g1) * *(g2++) * *(coef++);
		}
		g1++;
	    }
	    d += d0 * *(w++);
	}
    }
    else {
	phgError(1, "%s:%d: dimensions mismatch: coefficent A  %s (%d)",
		 __FILE__, __LINE__, DofTypeName(A), DofTypeDim(A));
    }

    return d * phgGeomGetVolume(u->g, e);
}

#if 0				/* note: use phgQuadCurlBasACurlBas */
/* computes integration of
 *        (curl of 'n'-th basis function of 'u')
 *        \cdot
 *        (curl of 'm'-th basis function of 'v')
 * on element 'e' using quadrature rule 'quad'.
 */
FLOAT
phgQuadCurlBasDotCurlBas(SIMPLEX *e, DOF *u, int n, DOF *v, int m, int order)
{
    int i, j, nvalues = DofTypeDim(u);
    FLOAT *g1, *g2, *w;
    FLOAT d, d0;
    QUAD *quad;

    assert(!SpecialDofType(u->type) && !SpecialDofType(v->type));

    if (nvalues != DofTypeDim(v))
	phgError(1, "%s:%d, dimensions mismatch: curl(%s) <==> curl(%s)\n",
		 __FILE__, __LINE__, u->name, v->name);

    if (order < 0)
	order = BasisOrder(u, e, n) - 1 + BasisOrder(v, e, m) - 1;
    quad = phgQuadGetQuad3D(order);

    g1 = phgQuadGetBasisCurl(e, u, n, quad);
    g2 = phgQuadGetBasisCurl(e, v, m, quad);

    d = 0.;
    w = quad->weights;
    for (i = 0; i < quad->npoints; i++) {
	d0 = 0.;
	for (j = 0; j < nvalues; j++) {
	    d0 += *(g1++) * (*(g2++));
	}
	d += d0 * (*(w++));
    }

    return d * phgGeomGetVolume(u->g, e);
}
#endif

FLOAT
phgQuadCurlBasACurlBas(SIMPLEX *e, DOF *u, int n, DOF *A, DOF *v, int m,
		       int order)
{
    int i, j, k, nvalues = DofTypeDim(u), An;
    const FLOAT *g1, *g2, *g2_tmp, *w, *coef;
    FLOAT d, d0;
    QUAD *quad;

    assert(!SpecialDofType(u->type) && !SpecialDofType(v->type));

    if (nvalues != DofTypeDim(v))
	phgError(1, "%s:%d, dimensions mismatch: curl(%s) <==> curl(%s)\n",
		 __FILE__, __LINE__, u->name, v->name);

    if (order < 0) {
	order = BasisOrder(u, e, n) - 1 + BasisOrder(v, e, m) - 1;
        if (A != NULL && DofTypeOrder(A, e) >= 0)
	    order += DofTypeOrder(A, e);
	if (order < 0)
	    order = 0;
    }
    quad = phgQuadGetQuad3D(order);

    g1 = phgQuadGetBasisCurl(e, u, n, quad);
    g2 = phgQuadGetBasisCurl(e, v, m, quad);
    d = 0.;
    w = quad->weights;
    if (NULL == A) {
	for (i = 0; i < quad->npoints; i++) {
	    d0 = 0.;
	    for (j = 0; j < nvalues; j++) {
		d0 += *(g1++) * (*(g2++));
	    }
	    d += d0 * (*(w++));
	}
	return d * phgGeomGetVolume(u->g, e);
    }

    An = DofDim(A);
    coef = phgQuadGetDofValues(e, A, quad);
    if (1 == An) {
	for (i = 0; i < quad->npoints; i++) {
	    d0 = 0.;
	    for (j = 0; j < nvalues; j++) {
		d0 += *(g1++) * *(g2++);
	    }
	    d += d0 * *(w++) * *(coef++);
	}
    }
    else if (An == nvalues) {
	for (i = 0; i < quad->npoints; i++) {
	    d0 = 0.;
	    for (j = 0; j < nvalues; j++) {
		d0 += *(g1++) * *(g2++) * *(coef++);
	    }
	    d += d0 * *(w++);
	}

    }
    else if (An == nvalues * nvalues) {
	for (i = 0; i < quad->npoints; i++) {
	    d0 = 0.;
	    g2_tmp = g2;
	    for (j = 0; j < nvalues; j++) {
		g2 = g2_tmp;
		for (k = 0; k < nvalues; k++) {
		    d0 += *(g1) * *(g2++) * *(coef++);
		}
		g1++;
	    }
	    d += d0 * *(w++);
	}
    }
    else {
	phgError(1, "%s:%d: dimensions mismatch: coefficent A  %s (%d)",
		 __FILE__, __LINE__, DofTypeName(A), DofTypeDim(A));
    }

    return d * phgGeomGetVolume(u->g, e);
}

FLOAT
phgQuadBasACurlBas(SIMPLEX *e, DOF *u, int n, DOF *A, DOF *v, int m,
		       int order)
{
    int i, j, k, nvalues = DofTypeDim(u), An;
    const FLOAT *g1, *g2, *g2_tmp, *w, *coef;
    FLOAT d, d0;
    QUAD *quad;

    assert(!SpecialDofType(u->type) && !SpecialDofType(v->type));

    if (nvalues != DofTypeDim(v))
	phgError(1, "%s:%d, dimensions mismatch: curl(%s) <==> curl(%s)\n",
		 __FILE__, __LINE__, u->name, v->name);

    if (order < 0) {
	order = BasisOrder(u, e, n) + BasisOrder(v, e, m) - 1;
        if (A != NULL && DofTypeOrder(A, e) >= 0)
	    order += DofTypeOrder(A, e);
	if (order < 0)
	    order = 0;
    }
    quad = phgQuadGetQuad3D(order);

    g1 = phgQuadGetBasisValues(e, u, n, quad);
    g2 = phgQuadGetBasisCurl(e, v, m, quad);
    d = 0.;
    w = quad->weights;
    if (NULL == A) {
	for (i = 0; i < quad->npoints; i++) {
	    d0 = 0.;
	    for (j = 0; j < nvalues; j++) {
		d0 += *(g1++) * (*(g2++));
	    }
	    d += d0 * (*(w++));
	}
	return d * phgGeomGetVolume(u->g, e);
    }

    An = DofDim(A);
    coef = phgQuadGetDofValues(e, A, quad);
    if (1 == An) {
	for (i = 0; i < quad->npoints; i++) {
	    d0 = 0.;
	    for (j = 0; j < nvalues; j++) {
		d0 += *(g1++) * *(g2++);
	    }
	    d += d0 * *(w++) * *(coef++);
	}
    }
    else if (An == nvalues) {
	for (i = 0; i < quad->npoints; i++) {
	    d0 = 0.;
	    for (j = 0; j < nvalues; j++) {
		d0 += *(g1++) * *(g2++) * *(coef++);
	    }
	    d += d0 * *(w++);
	}

    }
    else if (An == nvalues * nvalues) {
	for (i = 0; i < quad->npoints; i++) {
	    d0 = 0.;
	    g2_tmp = g2;
	    for (j = 0; j < nvalues; j++) {
		g2 = g2_tmp;
		for (k = 0; k < nvalues; k++) {
		    d0 += *(g1) * *(g2++) * *(coef++);
		}
		g1++;
	    }
	    d += d0 * *(w++);
	}
    }
    else {
	phgError(1, "%s:%d: dimensions mismatch: coefficent A  %s (%d)",
		 __FILE__, __LINE__, DofTypeName(A), DofTypeDim(A));
    }

    return d * phgGeomGetVolume(u->g, e);
}

FLOAT
phgQuadFuncDotBas(SIMPLEX *e, DOF_USER_FUNC userfunc, DOF *u, int n,
		  int order)
/* computes \int f * ('n'-th basis function of 'u') on element 'e'
 * using quadrature rule 'quad'. */
{
    int i, j, nvalues;
    const FLOAT *v, *bas, *funcvalues, *w;
    FLOAT d, d0;
    QUAD *quad;

    assert(!SpecialDofType(u->type));

    if (order < 0)
	order = DofTypeOrder(u, e) + BasisOrder(u, e, n);
    quad = phgQuadGetQuad3D(order);

    bas = phgQuadGetBasisValues(e, u, n, quad);
    funcvalues = phgQuadGetFuncValues(u->g, e, DofTypeDim(u), userfunc, quad);
    w = quad->weights;
    nvalues = DofTypeDim(u);

    d = 0.;
    for (v = funcvalues, i = 0; i < quad->npoints; i++) {
	d0 = 0.;
	for (j = 0; j < nvalues; j++) {
	    d0 += *(bas++) * (*(v++));
	}
	d += d0 * (*(w++));
    }

    return d * phgGeomGetVolume(u->g, e);
}

FLOAT
phgQuadDofDotBas(SIMPLEX *e, DOF *u, DOF *v, int n, int order)
/* computes \int 'u' * ('n'-th basis function of 'v') on element 'e'
 * using quadrature rule 'quad'. 
 */
{
    int i, j, nvalues;
    FLOAT d, d0;
    const FLOAT *bas, *dof, *w;
    QUAD *quad;

    assert(!SpecialDofType(v->type));

    if (order < 0) {
	i = DofTypeOrder(u, e);
	j = DofTypeOrder(v, e);
	if (i < 0)
	    i = j;
	order = i + j;
    }
    quad = phgQuadGetQuad3D(order);

    nvalues = DofDim(u);
    if(nvalues != DofTypeDim(v))
	phgError(1,"%s:%d: dimension mismatch\n", __FILE__, __LINE__);

    /* TODO: cache basis functions or values (analytic) of u */
    dof = phgQuadGetDofValues(e, u, quad);
    bas = phgQuadGetBasisValues(e, v, n, quad);
    w = quad->weights;
    d = 0.;
    for (i = 0; i < quad->npoints; i++) {
	d0 = 0.;
	for (j = 0; j < nvalues; j++) {
	    d0 += *(bas++) * *(dof++);
	}
	d += d0 * *(w++);
    }

    return d * phgGeomGetVolume(u->g, e);
}

FLOAT
phgQuadDofABas(SIMPLEX *e, DOF *u, DOF *A, DOF *v, int m, int order)
	/*
	 * computes integration of
	 *        DOF u 
	 *        \cdot
	 *        A coefficient
	 *        \cdot
	 *        ('m'-th basis function of 'v')
	 *      on element 'e' using quadrature rule 'quad'.
	 */
{
    int i, j, k, nvalues = DofDim(u), An;
    const FLOAT *g1, *g2, *g2_tmp, *coef, *w;
    FLOAT d, d0;
    QUAD *quad;

    assert(!SpecialDofType(u->type) && !SpecialDofType(v->type));
    assert(nvalues == DofTypeDim(v));

    if (order < 0) {
	order = ((j = DofTypeOrder(u, e)) >= 0 ? j : BasisOrder(v, e, m));
	order += BasisOrder(v, e, m);
        if (A != NULL && DofTypeOrder(A, e) >= 0)
	    order += DofTypeOrder(A, e);
    }
    quad = phgQuadGetQuad3D(order);

    d = 0.;
    g1 = phgQuadGetDofValues(e, u, quad);
    g2 = phgQuadGetBasisValues(e, v, m, quad);
    w = quad->weights;

    if (A == NULL) {
	for (i = 0; i < quad->npoints; i++) {
	    d0 = 0.;
	    for (j = 0; j < nvalues; j++) {
		d0 += *(g1++) * *(g2++);
	    }
	    d += d0 * (*(w++));
	}
	return d * phgGeomGetVolume(u->g, e);
    }

    An = DofDim(A);
    coef = phgQuadGetDofValues(e, A, quad);
    if (1 == An) {
	for (i = 0; i < quad->npoints; i++) {
	    d0 = 0.;
	    for (j = 0; j < nvalues; j++) {
		d0 += *(g1++) * *(g2++);
	    }
	    d += d0 * *(w++) * *(coef++);
	}
    }
    else if (An == nvalues) {
	for (i = 0; i < quad->npoints; i++) {
	    d0 = 0.;
	    for (j = 0; j < nvalues; j++) {
		d0 += *(g1++) * *(g2++) * *(coef++);
	    }
	    d += d0 * *(w++);
	}
    }
    else if (An == nvalues * nvalues) {
	for (i = 0; i < quad->npoints; i++) {
	    d0 = 0.;
	    g2_tmp = g2;
	    for (j = 0; j < nvalues; j++) {
		g2 = g2_tmp;
		for (k = 0; k < nvalues; k++)
		    d0 += *g1 * *(g2++) * *(coef++);
		g1++;
	    }
	    d += d0 * *(w++);
	}

    }
    else {
	phgError(1, "%s:%d: dimensions mismatch: coefficent A  %s (%d)",
		 __FILE__, __LINE__, DofTypeName(A), DofTypeDim(A));
    }

    return d * phgGeomGetVolume(u->g, e);
}

FLOAT
phgQuadDofAGradBas(SIMPLEX *e, DOF *u, DOF *A, DOF *v, int m, int order)
	/*
	 * computes integration of
	 *        DOF u 
	 *        \cdot
	 *        A coefficient
	 *        \cdot
	 *        (gradient of 'm'-th basis function of 'v')
	 *      on element 'e' using quadrature rule 'quad'.
	 */
{
    int i, j, k, nvalues = DofDim(u), An;
    const FLOAT *g1, *g2, *g2_tmp, *coef, *w;
    FLOAT d, d0;
    QUAD *quad;

    assert(!SpecialDofType(v->type));
    assert(nvalues == DofTypeDim(v) * Dim);

    if (order < 0) {
	order = ((j = DofTypeOrder(u, e)) >= 0 ? j : BasisOrder(v, e, m));
	order += BasisOrder(v, e, m) - 1;
        if (A != NULL && DofTypeOrder(A, e) >= 0)
	    order += DofTypeOrder(A, e);
	if (order < 0)
	    order = 0;
    }
    quad = phgQuadGetQuad3D(order);

    d = 0.;
    g1 = phgQuadGetDofValues(e, u, quad);
    g2 = phgQuadGetBasisGradient(e, v, m, quad);
    w = quad->weights;

    if (A == NULL) {
	for (i = 0; i < quad->npoints; i++) {
	    d0 = 0.;
	    for (j = 0; j < nvalues; j++) {
		d0 += *(g1++) * *(g2++);
	    }
	    d += d0 * (*(w++));
	}
	return d * phgGeomGetVolume(u->g, e);
    }

    An = DofDim(A);
    coef = phgQuadGetDofValues(e, A, quad);
    if (1 == An) {
	for (i = 0; i < quad->npoints; i++) {
	    d0 = 0.;
	    for (j = 0; j < nvalues; j++) {
		d0 += *(g1++) * *(g2++);
	    }
	    d += d0 * *(w++) * *(coef++);
	}
    }
    else if (An == nvalues) {
	for (i = 0; i < quad->npoints; i++) {
	    d0 = 0.;
	    for (j = 0; j < nvalues; j++) {
		d0 += *(g1++) * *(g2++) * *(coef++);
	    }
	    d += d0 * *(w++);
	}
    }
    else if (An == nvalues * nvalues) {
	for (i = 0; i < quad->npoints; i++) {
	    d0 = 0.;
	    g2_tmp = g2;
	    for (j = 0; j < nvalues; j++) {
		g2 = g2_tmp;
		for (k = 0; k < nvalues; k++)
		    d0 += *g1 * *(g2++) * *(coef++);
		g1++;
	    }
	    d += d0 * *(w++);
	}

    }
    else {
	phgError(1, "%s:%d: dimensions mismatch: coefficent A  %s (%d)",
		 __FILE__, __LINE__, DofTypeName(A), DofTypeDim(A));
    }

    return d * phgGeomGetVolume(u->g, e);
}

FLOAT *
phgQuadDofDotCurlBas_(SIMPLEX *e, DOF *u, DOF *v, int m, int order, FLOAT *res)
{
    int i,  nvalues = DofDim(u);
    const FLOAT *g1, *g2, *w;
    FLOAT d;
    QUAD *quad;

    assert(!SpecialDofType(v->type));

    if (order < 0) {
	order = BasisOrder(v, e, m) - 1;
	if (DofTypeOrder(u, e) >= 0)
	    order += DofTypeOrder(u, e);
	if (order < 0)
	    order = 0;
    }
    quad = phgQuadGetQuad3D(order);

    g1 = phgQuadGetDofValues(e, u, quad);
    g2 = phgQuadGetBasisGradient(e, v, m, quad);
    w = quad->weights;

    if (nvalues == Dim && v->dim == Dim && DofTypeDim(v) == 1) {
        res[0] = res[1] = res[2] = 0.; 
	for (i = 0; i < quad->npoints; i++) {
	    res[0] += (g1[1] * g2[2] - g1[2] * g2[1]) * (*w);
	    res[1] += (g1[2] * g2[0] - g1[0] * g2[2]) * (*w);
	    res[2] += (g1[0] * g2[1] - g1[1] * g2[0]) * (*w);
	    w++;
            g2 += Dim;
	}
	d = phgGeomGetVolume(u->g, e);
        res[0] *= d;
        res[1] *= d;
        res[2] *= d;
	return res;
    } else {
	phgError(1, "%s:%d: unimplement  ", __FILE__, __LINE__);
        return NULL;
    }
}

FLOAT
phgQuadDofACurlBas(SIMPLEX *e, DOF *u, DOF *A, DOF *v, int m, int order)
	/*
	 * computes integration of
	 *        DOF u 
	 *        \cdot
	 *        A coefficient
	 *        \cdot
	 *        (Curl of 'm'-th basis function of 'v')
	 *      on element 'e' using quadrature rule 'quad'.
	 */
{
    int i, j, k, nvalues = DofDim(u), An;
    const FLOAT *g1, *g2, *g2_tmp, *coef, *w;
    FLOAT d, d0;
    QUAD *quad;

    assert(!SpecialDofType(v->type));
    assert(nvalues == DofTypeDim(v));

    if (order < 0) {
	order = ((j = DofTypeOrder(u, e)) >= 0 ? j : BasisOrder(v, e, m));
	order += BasisOrder(v, e, m) - 1;
	if (A != NULL && DofTypeOrder(A, e) >= 0)
	    order += DofTypeOrder(A, e);
	if (order < 0)
	    order = 0;
    }
    quad = phgQuadGetQuad3D(order);

    d = 0.;
    g1 = phgQuadGetDofValues(e, u, quad);
    g2 = phgQuadGetBasisCurl(e, v, m, quad);
    w = quad->weights;

    if (A == NULL) {
	for (i = 0; i < quad->npoints; i++) {
	    d0 = 0.;
	    for (j = 0; j < nvalues; j++) {
		d0 += *(g1++) * *(g2++);
	    }
	    d += d0 * *(w++);
	}
	return d * phgGeomGetVolume(u->g, e);
    }

    An = DofDim(A);
    coef = phgQuadGetDofValues(e, A, quad);
    if (1 == An) {
	for (i = 0; i < quad->npoints; i++) {
	    d0 = 0.;
	    for (j = 0; j < nvalues; j++) {
		d0 += *(g1++) * *(g2++);
	    }
	    d += d0 * *(w++) * *(coef++);
	}
    }
    else if (An == nvalues) {
	for (i = 0; i < quad->npoints; i++) {
	    d0 = 0.;
	    for (j = 0; j < nvalues; j++) {
		d0 += *(g1++) * *(g2++) * *(coef++);
	    }
	    d += d0 * *(w++);
	}
    }
    else if (An == nvalues * nvalues) {
	for (i = 0; i < quad->npoints; i++) {
	    d0 = 0.;
	    g2_tmp = g2;
	    for (j = 0; j < nvalues; j++) {
		g2 = g2_tmp;
		for (k = 0; k < nvalues; k++)
		    d0 += *g1 * *(g2++) * *(coef++);
		g1++;
	    }
	    d += d0 * *(w++);
	}
    }
    else {
	phgError(1, "%s:%d: dimensions mismatch: coefficent A  %s (%d)",
		 __FILE__, __LINE__, DofTypeName(A), DofTypeDim(A));
    }

    return d * phgGeomGetVolume(u->g, e);
}

FLOAT
phgQuadGradBasDotBas(SIMPLEX *e, DOF *s, int m, DOF *v, int n, int order)
	/*
	   S is the voltage soure
	   We choose it as the base function of P1 at vertex m,
	   this function do 
	   int
	   grad S * v
	 */
{
    int i, j, nvalues = DofTypeDim(v);
    const FLOAT *g1, *g2, *w, *lambda;
    FLOAT d, d0;
    QUAD *quad;

    assert(!SpecialDofType(s->type) && !SpecialDofType(v->type));

    if (nvalues != Dim * DofTypeDim(s))
	phgError(1, "%s:%d, dimensions mismatch: grad(%s) <==> (%s)\n",
		 __FILE__, __LINE__, s->name, v->name);

    if (order < 0)
	order = BasisOrder(s, e, m) - 1 + BasisOrder(v, e, n);
    if (order < 0)
	order = 0;
    quad = phgQuadGetQuad3D(order);

    g1 = phgQuadGetBasisGradient(e, s, m, quad);
    g2 = phgQuadGetBasisValues(e, v, n, quad);
    d = 0.;
    lambda = quad->points;
    w = quad->weights;
    for (i = 0; i < quad->npoints; i++) {
	d0 = 0.;
	for (j = 0; j < nvalues; j++) {
	    d0 += *(g1++) * (*(g2++));
	}
	d += d0 * (*(w++));
	lambda += Dim + 1;
    }
    return d * phgGeomGetVolume(s->g, e);
}
