/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2009 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: solver-gather.c,v 1.14 2009/05/28 01:37:30 zlb Exp $
 *
 * This solver gathers all (non diagonal) rows of the system to process 0 and
 * calls a serial solver to solve it on process 0 */

#include "phg.h"

#define Initialize		NULL
#define Finalize		NULL
#define AddMatrixEntries	NULL
#define AddRHSEntries		NULL

#include <stdlib.h>
#include <string.h>

typedef struct {
    SOLVER	*solver;
    INT		*index;		/* unpacked=>packed map */
    FLOAT	*idiag;		/* inversed diagonal */
    MAT		*mat;		/* saved columns */
    INT		N, n;		/* global/local size of the gathered system */

    int		*cnts, *dsps;
    char	*solver_opts;
    int		solver_index;
    BOOLEAN	factored;
} OEM_DATA;

#define	_t	((OEM_DATA *)solver->oem_data)

static int solver_index = 0;
static char *solver_opts = NULL;

static int
RegisterOptions(void)
{
    /* set default solver to the first direct solver (if available) */
    for (solver_index = 0;
	 phgSolverNames[solver_index] != NULL
	    && (phgSolverList[solver_index] == SOLVER_GATHER ||
		phgSolverList[solver_index] == SOLVER_BUILTIN ||
		phgSolverList[solver_index]->iterative == TRUE);
	 solver_index++);
    if (phgSolverNames[solver_index] == NULL)
	 solver_index = 0;	/* fall back to the first solver */

    phgOptionsRegisterTitle("\nThe Gather solver options:", NULL, "gather");
    phgOptionsRegisterKeyword("-gather_solver", "The serial solver used to "
					"solve the gathered system",
					phgSolverNames, &solver_index);
    phgOptionsRegisterString("-gather_solver_opts",
					"Options passed to the serial solver",
					&solver_opts);

    return 0;
}

static int
Init(SOLVER *solver)
{
    solver->oem_data = phgCalloc(1, sizeof(OEM_DATA));
    _t->solver_index = solver_index;
    if (solver_opts != NULL)
	_t->solver_opts = strdup(solver_opts);

    return 0;
}

static int
Create(SOLVER *solver)
{
    return 0;
}

static int
Destroy(SOLVER *solver)
{
    if (solver->oem_data == NULL)
	return 0;

    phgFree(_t->solver_opts);
    phgFree(_t->index);
    phgFree(_t->idiag);
    phgFree(_t->cnts);
    phgFree(_t->dsps);
    phgMatDestroy(&_t->mat);
    phgSolverDestroy(&_t->solver);

    phgFree(solver->oem_data);
    solver->oem_data = NULL;

    return 0;
}

static int
Assemble(SOLVER *solver)
{
    int rank, nprocs, n0;
    INT nlocal, i, j, k, n;
    MAP *map;
    MAT *B;
    OEM_SOLVER *oem_solver;
    const MAT_ROW *row;
    MAT_ROW *row1;
    FLOAT *data0, *data1;
    INT alloc, ncols0, *cols0, ncols1, *cols1;
    double time0 = 0.;

   /* Note: currently only ordinary matrices allowed since it's easier and
    *	    more efficient to work with local column indices. */
    if (solver->mat->type == PHG_PACKED)
	phgMatUnpack(solver->mat);
    assert(solver->mat->type == PHG_UNPACKED);

    if (_t->factored)
	return 0;

    _t->factored = TRUE;

    if (solver->sent_to_oem == TRUE)
	phgError(1, "%s:%d, data already sent to OEM solver!\n",
						__FILE__, __LINE__);
    solver->sent_to_oem = TRUE;

    rank = solver->mat->rmap->rank;
    nprocs = solver->mat->rmap->nprocs;
    nlocal = solver->mat->rmap->nlocal;
    _t->index = phgAlloc(solver->mat->localsize * sizeof(*_t->index));
    _t->idiag = phgAlloc(solver->mat->localsize * sizeof(*_t->idiag));

    /* Count # of non-diagonal rows on each process. */
    if (phgVerbosity > 0)
	time0 = phgGetTime(NULL);
    n = 0;
    for (i = 0; i < nlocal; i++) {
	/* count diagonal and non-diagonal rows */
	/*row = phgMatGetRow(solver->mat, i);*/
	row = solver->mat->rows + i;
	if (row->ncols <= 1) {
	    if (row->ncols == 0. || row->data[0] == 0.)
		phgError(1, "(%s:%d) row %d is empty, abort.\n",
			 i + solver->mat->rmap->partition[rank],
			 __FILE__, __LINE__);
	    _t->idiag[i] = 1. / row->data[0];
	    _t->index[i] = -1;
	    if (solver->mat->blocks == NULL && solver->mat->refcount == 0) {
		/* free the matrix row */
		row1 = solver->mat->rows + i;
		phgFree(row1->cols);
		phgFree(row1->data);
		row1->cols = NULL;
		row1->data = NULL;
		row1->ncols = row1->alloc = 0;
	    }
	}
	else {
	    _t->idiag[i] = 0.0;
	    _t->index[i] = n++;
	}
    }
    _t->n = n;

#if USE_MPI
    _t->cnts = phgAlloc(nprocs * sizeof(*_t->cnts));
    _t->dsps = phgAlloc(nprocs * sizeof(*_t->dsps));
    n0 = n;
    MPI_Gather(&n0, 1, MPI_INT, _t->cnts, 1, MPI_INT, 0,
					solver->mat->rmap->comm);
    _t->N = 0;
    if (rank == 0) {
	for (n0 = 0; n0 < nprocs; n0++) {
	    _t->dsps[n0] = _t->N;
	    _t->N += _t->cnts[n0];
	}
    }
    MPI_Bcast(&_t->N, 1, PHG_MPI_INT, 0, solver->mat->rmap->comm);
    MPI_Scatter(_t->dsps, 1, MPI_INT, &n0, 1, MPI_INT, 0,
					solver->mat->rmap->comm);

    if (nprocs > 1 && solver->mat->cinfo == NULL)
	    solver->mat->cinfo = phgMapCreateCommInfo(
					solver->mat->cmap->comm,
					nprocs,
					rank,
					nlocal,
					solver->mat->localsize,
					solver->mat->O2Gmap,
					solver->mat->ordering,
					solver->mat->cmap->partition);

    if (nprocs > 1 && solver->mat->cinfo != NULL) {
	/* get new numbering of off-process entries.
	 * Hacky: use phgMapScatter() to exchange flags */
	FLOAT *data, *offp_data;
	data = phgAlloc(nlocal * sizeof(*data));
	offp_data = phgAlloc(solver->mat->cinfo->rsize * sizeof(*offp_data));
	for (i = 0; i < nlocal; i++) {
	    if (_t->index[i] < 0)
		data[i] = -2.0;
	    else
		data[i] = (FLOAT)(_t->index[i] + n0);
	}
	phgMapScatter(solver->mat->cinfo, 1, data, offp_data);
	phgFree(data);
	for (i = nlocal; i < solver->mat->localsize; i++) {
	    j = (INT)(offp_data[i - nlocal] + 0.5);
	    _t->index[i] = j;
	    assert(j < _t->N);
	}
	phgFree(offp_data);
    }
#else	/* USE_MPI */
    n0 = 0;
    _t->N = n;
#endif	/* USE_MPI */

    if (phgVerbosity > 0) {
	double time1 = phgGetTime(NULL);
	phgPrintf("Determining derived system: size = %d, wtime = %0.4lg\n",
			_t->N, time1 - time0);
	time0 = time1;
    }

    /* Matrix containing saved columns */
    map = phgMapCreateSimpleMap(solver->mat->rmap->comm, _t->n, _t->N);
    _t->mat = phgMapCreateMat(map, solver->mat->cmap);
    _t->mat->handle_bdry_eqns = FALSE;
    phgMapDestroy(&map);

    /* Temperary matrix for gathering rows to process 0 */
    map = phgMapCreateSimpleMap(solver->mat->rmap->comm,
					rank == 0 ? _t->N : 0, _t->N);
    B = phgMapCreateMat(map, map);
    B->handle_bdry_eqns = FALSE;
    phgMapDestroy(&map);

    /* send rows to B and _t->mat */
    data0 = data1 = NULL;
    cols0 = cols1 = NULL;
    alloc = 0;
    for (i = 0; i < nlocal; i++) {
	if (_t->index[i] < 0)
	    continue;
	/*row = phgMatGetRow(solver->mat, i);*/
	row = solver->mat->rows + i;
	if (alloc < row->ncols) {
	    alloc = row->ncols;
	    phgFree(data0);
	    data0 = phgAlloc(2 * alloc * sizeof(*data0));
	    data1 = data0 + alloc;
	    phgFree(cols0);
	    cols0 = phgAlloc(2 * alloc * sizeof(*cols0));
	    cols1 = cols0 + alloc;
	}
	ncols0 = ncols1 = 0;
	for (j = 0; j < row->ncols; j++) {
	    k = row->cols[j];
	    if (_t->index[k] >= 0) {
		data0[ncols0] = row->data[j];
		if (k < nlocal)
		    cols0[ncols0] = _t->index[k] + n0;
		else
		    cols0[ncols0] = _t->index[k];
		ncols0++;
	    }
	    else {
		data1[ncols1] = row->data[j];
#if USE_MPI
		if (k >= nlocal)
		    cols1[ncols1] = solver->mat->O2Gmap[k - nlocal];
		else
#endif	/* USE_MPI */
		    cols1[ncols1] = k + solver->mat->cmap->partition[rank];
		ncols1++;
	    }
	}
	if (solver->mat->blocks == NULL && solver->mat->refcount == 0) {
	    /* free the matrix row */
	    row1 = solver->mat->rows + i;
	    phgFree(row1->cols);
	    phgFree(row1->data);
	    row1->cols = NULL;
	    row1->data = NULL;
	    row1->ncols = row1->alloc = 0;
	}
	j = _t->index[i] + n0;
	phgMatAddGlobalEntries(B, 1, &j, ncols0, cols0, data0);
	phgMatAddLGEntries(_t->mat, 1, _t->index + i, ncols1, cols1, data1);
    }
    phgFree(data0);
    phgFree(cols0);
    phgMatAssemble(B);
    phgMatAssemble(_t->mat);

    if (phgVerbosity > 0) {
	time0 = phgGetTime(NULL) - time0;
	phgPrintf("Setting up derived system: wtime = %0.4lg\n", time0);
    }

    if (rank > 0) {
	phgMatDestroy(&B);
	return 0;
    }

    /* Hacky: replace B->rmap and B->cmap with the serial map */
    phgMapDestroy(&B->rmap);
    phgMapDestroy(&B->cmap);
    B->rmap = phgMapCreateSimpleMap(MPI_COMM_SELF, _t->N, _t->N);
    B->cmap = phgMatGetRowMap(B);	/* this increments B->cmap->refcount */
#if USE_MPI
    if (B->cinfo != NULL)
	phgMapDestroyCommInfo(&B->cinfo);
#endif	/* USE_MPI */

    /* Create the serial solver on process 0.
     *
     * Note: it's better to create the solver in 'Init()' to honor all solver
     * options for the serial solver, but we couldn't do it there since we
     * need actual matrix data to determine the serial map. */

    phgOptionsPush();
    /* set options for the serial solver */
    phgOptionsSetOptions(_t->solver_opts);
    oem_solver = phgSolverList[_t->solver_index];
    if (oem_solver == solver->oem_solver)
	phgError(1, "(%s:%d) solver regression detected, abort.\n",
						__FILE__, __LINE__);
    assert(_t->solver == NULL);
    _t->solver = phgMat2Solver(oem_solver, B);
    phgOptionsPop();
    phgMatDestroy(&B);

    return 0;
}
 
static int
Solve(SOLVER *solver, VEC *x, BOOLEAN destroy)
{
    INT i, j;
    VEC *v;
    FLOAT *buffer0, *buffer1;
    int rank = solver->mat->rmap->rank;
    double time0 = 0.;

    assert(x->nvec == 1);
    Assemble(solver);

    if (phgVerbosity > 0)
	time0 = phgGetTime(NULL);

    /* copy initial solution and updated RHS to buffer0 */
    buffer0 = phgAlloc(2 * _t->n * sizeof(*buffer0));
    for (i = j = 0; i < solver->mat->rmap->nlocal; i++) {
	if (_t->index[i] >= 0) {
	    buffer0[2 * j + 1] = x->data[i];
	    x->data[i] = 0.0;
	    j++;
	    continue;
	}
	assert(_t->idiag[i] != 0.0);
	x->data[i] = solver->rhs->data[i] * _t->idiag[i];
    }
    v = phgMatVec(MAT_OP_N, 1.0, _t->mat, x, 0.0, NULL);
    for (i = j = 0; i < solver->mat->rmap->nlocal; i++) {
	if (_t->index[i] < 0)
	    continue;
	buffer0[2 * j] = solver->rhs->data[i] - v->data[j];
	j++;
    }
    phgVecDestroy(&v);

#if USE_MPI
    if (solver->rhs->map->nprocs > 1) {
	MPI_Datatype type;
	buffer1 = (rank == 0 ? phgAlloc(2 * _t->N * sizeof(*buffer1)) : NULL);
	MPI_Type_contiguous(2, PHG_MPI_FLOAT, &type);
	MPI_Type_commit(&type);
	MPI_Gatherv(buffer0, _t->n, type, buffer1, _t->cnts, _t->dsps, type,
                0, solver->mat->rmap->comm);
	MPI_Type_free(&type);
    }
    else
#endif	/* use_MPI */
    {
	buffer1 = buffer0;
    }

    if (phgVerbosity > 0) {
	double time1 = phgGetTime(NULL);
	phgPrintf("Setting up initial solution and RHS: wtime = %0.4lg\n",
			time1 - time0);
	time0 = time1;
    }

    if (rank == 0) {
	INT verb = phgVerbosity;
	VEC *x0 = phgMapCreateVec(_t->solver->rhs->map, 0);
	x0->nvec = 1;
	x0->data = buffer1;
	for (i = 0; i < _t->N; i++) {
	    _t->solver->rhs->data[i] = buffer1[2 * i];
	    buffer1[i] = buffer1[2 * i + 1];
	}
	phgSolverAssemble(_t->solver);
	if (phgVerbosity > 0)
	    phgVerbosity--;
	phgSolverVecSolve(_t->solver, destroy, x0);
	phgVerbosity = verb;
	x0->data = NULL;
	phgVecDestroy(&x0);
    }

    if (phgVerbosity > 0) {
	double time1 = phgGetTime(NULL);
	phgPrintf("Solving derived system: wtime = %0.4lg\n", time1 - time0);
	time0 = time1;
    }

#if USE_MPI
    if (solver->rhs->map->nprocs > 1) {
	MPI_Scatterv(buffer1, _t->cnts, _t->dsps, PHG_MPI_FLOAT,
		     buffer0, _t->n, PHG_MPI_FLOAT, 0, solver->mat->rmap->comm);
	phgFree(buffer1);
    }
#endif	/* use_MPI */

    for (i = j = 0; i < solver->mat->rmap->nlocal; i++) {
	if (_t->index[i] < 0)
	    continue;
	x->data[i] = buffer0[j++];
    }

    phgFree(buffer0);

    if (rank == 0) {
	solver->residual = _t->solver->residual;
	solver->nits = _t->solver->nits;
    }
#if USE_MPI
    {
	double a[2];
	a[0] = solver->residual;
	a[1] = solver->nits;
	MPI_Bcast(a, 2, MPI_DOUBLE, 0, solver->mat->rmap->comm);
	solver->residual = a[0];
	solver->nits = (INT)(a[1] + 0.5);
    }
#endif	/* use_MPI */

    if (destroy) {
	phgFree(_t->solver_opts);
	_t->solver_opts = NULL;
	phgFree(_t->index);
	_t->index = NULL;
	phgFree(_t->idiag);
	_t->idiag = NULL;
	phgFree(_t->cnts);
	_t->cnts = NULL;
	phgFree(_t->dsps);
	_t->dsps = NULL;
	phgMatDestroy(&_t->mat);
	phgSolverDestroy(&_t->solver);
    }

    if (phgVerbosity > 0) {
	double time1 = phgGetTime(NULL);
	phgPrintf("Scattering solution: wtime = %0.4lg\n", time1 - time0);
	time0 = time1;
    }

    return solver->nits;
}

/*--------------------------------------------------------------------*/

OEM_SOLVER phgSolverGather_ = {
    "Gather", RegisterOptions,
    Initialize, Finalize, Init, Create, Destroy,
    AddMatrixEntries, AddRHSEntries,
    Assemble, Solve, NULL, NULL, NULL, TRUE, TRUE
};
